"""`OffshoreSubstationInstallation` and related processes."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


import simpy
from marmot import process
from wisdem.orbit.core import Vessel
from wisdem.orbit.core.logic import shuttle_items_to_queue, prep_for_site_operations
from wisdem.orbit.phases.install import InstallPhase
from wisdem.orbit.phases.install.monopile_install.common import Monopile, upend_monopile, install_monopile

from .common import Topside, install_topside


class OffshoreSubstationInstallation(InstallPhase):
    """
    Offshore Substation (OSS) installation process using a single heavy lift
    vessel and feeder barge.
    """

    phase = "Offshore Substation Installation"
    capex_category = "Offshore Substation"

    #:
    expected_config = {
        "num_substations": "int",
        "oss_install_vessel": "dict | str",
        "num_feeders": "int (optional, default: 1)",
        "feeder": "dict | str",
        "site": {"distance": "km", "depth": "m"},
        "port": {
            "num_cranes": "int (optional, default: 1)",
            "monthly_rate": "USD/mo (optional)",
            "name": "str (optional)",
        },
        "offshore_substation_topside": {
            "deck_space": "m2",
            "mass": "t",
            "unit_cost": "USD",
        },
        "offshore_substation_substructure": {
            "type": "Monopile",
            "deck_space": "m2",
            "mass": "t",
            "length": "m",
            "unit_cost": "USD",
        },
    }

    def __init__(self, config, weather=None, **kwargs):
        """
        Creates an instance of OffshoreSubstationInstallation.

        Parameters
        ----------
        config : dict
            Simulation specific configuration.
        weather : np.ndarray
            Weather profile at site.
        """

        super().__init__(weather, **kwargs)

        config = self.initialize_library(config, **kwargs)
        self.config = self.validate_config(config)

        self.initialize_port()
        self.setup_simulation(**kwargs)

    @property
    def system_capex(self):
        """Returns procurement CapEx of the offshore substations."""

        return self.config["num_substations"] * (
            self.config["offshore_substation_topside"]["unit_cost"]
            + self.config["offshore_substation_substructure"]["unit_cost"]
        )

    def setup_simulation(self, **kwargs):
        """
        Initializes required objects for simulation.
        - Creates port + crane
        - Creates monopile and topside
        - Creates heavy lift vessel and feeder
        """

        self.initialize_topsides_and_substructures()
        self.initialize_oss_install_vessel()
        self.initialize_feeders()
        self.initialize_queue()

        site_distance = self.config["site"]["distance"]
        site_depth = self.config["site"]["depth"]
        num_subsations = self.config["num_substations"]

        install_oss_from_queue(
            self.oss_vessel,
            queue=self.active_feeder,
            site_depth=site_depth,
            distance=site_distance,
            substations=num_subsations,
            **kwargs,
        )

        component_list = ["Monopile", "Topside"]

        for feeder in self.feeders:
            shuttle_items_to_queue(
                feeder,
                port=self.port,
                queue=self.active_feeder,
                distance=site_distance,
                items=component_list,
                **kwargs,
            )

    def initialize_topsides_and_substructures(self):
        """
        Creates offshore substation objects at port.
        """

        top = Topside(**self.config["offshore_substation_topside"])
        sub = Monopile(**self.config["offshore_substation_substructure"])
        self.num_substations = self.config["num_substations"]

        for _ in range(self.num_substations):
            self.port.put(sub)
            self.port.put(top)

    def initialize_oss_install_vessel(self):
        """
        Creates the offshore substation installation vessel object.
        """

        oss_vessel_specs = self.config.get("oss_install_vessel", None)
        name = oss_vessel_specs.get("name", "Heavy Lift Vessel")

        oss_vessel = self.initialize_vessel(name, oss_vessel_specs)
        self.env.register(oss_vessel)

        oss_vessel.initialize()
        oss_vessel.at_port = True
        oss_vessel.at_site = False
        self.oss_vessel = oss_vessel

    def initialize_feeders(self):
        """
        Initializes feeder barge objects.
        """

        number = self.config.get("num_feeders", 1)
        feeder_specs = self.config.get("feeder", None)

        self.feeders = []
        for n in range(number):
            # TODO: Add in option for named feeders.
            name = "Feeder {}".format(n)

            feeder = self.initialize_vessel(name, feeder_specs)
            self.env.register(feeder)

            feeder.initialize()
            feeder.at_port = True
            feeder.at_site = False
            self.feeders.append(feeder)

    def initialize_queue(self):
        """
        Creates the queue that feeders will join at site. Limited to one active
        feeder at a time.
        """

        self.active_feeder = simpy.Resource(self.env, capacity=1)
        self.active_feeder.vessel = None
        self.active_feeder.activate = self.env.event()

    @property
    def detailed_output(self):
        """Returns detailed outputs of the oss installation."""

        if self.feeders:
            transport_vessels = [*self.feeders]

        else:
            transport_vessels = [self.oss_vessel]

        outputs = {
            self.phase: {
                **self.agent_efficiencies,
                **self.get_max_cargo_mass_utilzations(transport_vessels),
                **self.get_max_deck_space_utilzations(transport_vessels),
            }
        }

        return outputs


@process
def install_oss_from_queue(vessel, queue, substations, distance, **kwargs):
    """
    Installs offshore subsations and substructures from queue of feeder barges.

    Parameters
    ----------
    env : Environment
    vessel : Vessel
    queue : simpy.Resource
        Queue object to shuttle to.
    number : int
        Number of substructures to install.
    distance : int | float
        Distance from site to port (km).
    """

    n = 0
    while n < substations:
        if vessel.at_port:
            # Transit to site
            vessel.at_port = False
            yield vessel.transit(distance)
            vessel.at_site = True

        if vessel.at_site:

            if queue.vessel:

                # Prep for monopile install
                yield prep_for_site_operations(vessel, survey_required=True, **kwargs)

                # Get monopile
                monopile = yield vessel.get_item_from_storage("Monopile", vessel=queue.vessel, **kwargs)

                yield upend_monopile(vessel, monopile.length, **kwargs)
                yield install_monopile(vessel, monopile, **kwargs)

                # Get topside
                topside = yield vessel.get_item_from_storage("Topside", vessel=queue.vessel, release=True, **kwargs)
                yield install_topside(vessel, topside, **kwargs)
                n += 1

            else:
                start = vessel.env.now
                yield queue.activate
                delay_time = vessel.env.now - start
                vessel.submit_action_log("Delay", delay_time, location="Site")

    # Transit to port
    vessel.at_site = False
    yield vessel.transit(distance)
    vessel.at_port = True
    vessel.submit_debug_log(
        message="Monopile installation complete!",
        progress="Offshore Substation",
    )



from __future__ import print_function
import numpy as np
from wisdem.commonse.utilities import nodal2sectional
import openmdao.api as om


class CylindricalShellProperties(om.ExplicitComponent):
    """
    OpenMDAO wrapper for tube class to obtain cylindrical sheel properties.

    Parameters
    ----------
    d : numpy array[nFull], [m]
        tower diameter at corresponding locations
    t : numpy array[nFull-1], [m]
        shell thickness at corresponding locations

    Returns
    -------
    Az : numpy array[nFull-1], [m**2]
        cross-sectional area
    Asx : numpy array[nFull-1], [m**2]
        x shear area
    Asy : numpy array[nFull-1], [m**2]
        y shear area
    Jz : numpy array[nFull-1], [m**4]
        polar moment of inertia
    Ixx : numpy array[nFull-1], [m**4]
        area moment of inertia about x-axis
    Iyy : numpy array[nFull-1], [m**4]
        area moment of inertia about y-axis

    """

    def initialize(self):
        self.options.declare("nFull")

    def setup(self):
        nFull = self.options["nFull"]

        self.add_input("d", np.zeros(nFull), units="m")
        self.add_input("t", np.zeros(nFull - 1), units="m")

        self.add_output("Az", np.zeros(nFull - 1), units="m**2")
        self.add_output("Asx", np.zeros(nFull - 1), units="m**2")
        self.add_output("Asy", np.zeros(nFull - 1), units="m**2")
        self.add_output("Jz", np.zeros(nFull - 1), units="m**4")
        self.add_output("Ixx", np.zeros(nFull - 1), units="m**4")
        self.add_output("Iyy", np.zeros(nFull - 1), units="m**4")

        # Derivatives
        self.declare_partials("*", "*", method="fd", form="central", step=1e-6)

    def compute(self, inputs, outputs):
        d, _ = nodal2sectional(inputs["d"])
        tube = Tube(d, inputs["t"])

        outputs["Az"] = tube.Area
        outputs["Asx"] = tube.Asx
        outputs["Asy"] = tube.Asy
        outputs["Jz"] = tube.J0
        outputs["Ixx"] = tube.Jxx
        outputs["Iyy"] = tube.Jyy


class Tube:
    """The Tube Class contains functions to calculate properties of tubular circular cross-sections
    for structural analyses."""

    def __init__(self, D, t, Lgth=np.NaN, Kbuck=1.0):
        self.D = D
        self.t = t
        self.L = Lgth * np.ones(np.size(D))  # this makes sure we exapnd Lght if D,t, arrays
        self.Kbuck = Kbuck * np.ones(np.size(D))  # this makes sure we exapnd Kbuck if D,t, arrays

    @property
    def Area(self):  # Cross sectional area of tube
        return (self.D ** 2 - (self.D - 2 * self.t) ** 2) * np.pi / 4

    @property
    def derivArea(self):
        return {"D": 2 * np.pi / 4 * (self.D ** 2 - (self.D - 2 * self.t)) * (2 * self.D - 1), "t": 0}

    @property
    def Amid(self):  # mid-thickness inscribed area of tube (thin wall torsion calculation)
        return (self.D - self.t) ** 2 * np.pi / 4

    @property
    def Jxx(self):  # 2nd area moment of inertia w.r.t. x-x axis (Jxx=Jyy for tube)
        return (self.D ** 4 - (self.D - 2 * self.t) ** 4) * np.pi / 64

    @property
    def Jyy(self):  # 2nd area moment of inertia w.r.t. x-x axis (Jxx=Jyy for tube)
        return self.Jxx

    @property
    def J0(self):  # polar moment of inertia w.r.t. z-z axis (torsional)
        return 2.0 * self.Jxx

    @property
    def Asy(self):  # Shear Area for tubular cross-section
        Ri = self.D / 2 - self.t
        Ro = self.D / 2
        return self.Area / (1.124235 + 0.055610 * (Ri / Ro) + 1.097134 * (Ri / Ro) ** 2 - 0.630057 * (Ri / Ro) ** 3)

    @property
    def Asx(self):  # Shear Area for tubular cross-section
        return self.Asy

    @property
    def BdgMxx(self):  # Bending modulus for tubular cross-section
        return self.Jxx / (self.D / 2)

    @property
    def BdgMyy(self):  # Bending modulus for tubular cross-section =BdgMxx
        return self.Jyy / (self.D / 2)

    @property
    def TorsConst(self):  # Torsion shear constant for tubular cross-section
        return self.J0 / (self.D / 2)

    @property
    def S(self):  # Bending modulus for tubular cross-section
        return self.BdgMxx

    @property
    def C(self):  # Torsion shear constant for tubular cross-section
        return self.TorsConst

    @property
    def Rgyr(self):  # Radius of Gyration for circular tube
        return np.sqrt(self.Jxx / self.Area)

    @property
    def Klr(self):  # Klr buckling parameter
        return self.Kbuck * self.L / self.Rgyr


class IBeam:
    def __init__(self, L_flange, t_flange, H_web, t_web):
        self.Lf = L_flange
        self.tf = t_flange
        self.Hw = H_web
        self.tw = t_web
        self.H = H_web + 2 * t_flange

    @property
    def AreaFlange(self):  # Cross sectional area of tube
        return self.Lf * self.tf

    @property
    def AreaWeb(self):  # Cross sectional area of tube
        return self.Hw * self.tw

    @property
    def Area(self):  # Cross sectional area of tube
        return self.AreaWeb + 2 * self.AreaFlange

    @property
    def Iyy(self):  # 2nd area moment of inertia w.r.t. y-y axis running parallel to flange through CG
        return (self.Lf * self.H ** 3 - (self.Lf - self.tw) * self.Hw ** 3) / 12.0

    @property
    def Izz(self):  # 2nd area moment of inertia w.r.t. z-z running through center of web
        return (2 * self.tw * self.Lf ** 3 + self.Hw * self.tw ** 3) / 12.0

    @property
    def Jxx(self):  # polar moment of inertia w.r.t. z-z axis (torsional)
        return 2 * self.Lf * self.tf ** 3 + self.H * self.tw ** 3

    @property
    def Asy(self):  # Shear Area for tubular cross-section
        return 1.64 * self.Lf * self.tf

    @property
    def Asz(self):  # Shear Area for tubular cross-section
        return self.tw * self.H

    @property
    def BdgMyy(self):  # Bending modulus for tubular cross-section
        return 2 * self.Iyy / self.H

    @property
    def BdgMzz(self):  # Bending modulus for tubular cross-section =BdgMxx
        return 2 * self.Izz / self.Lf

    @property
    def TorsConst(self):  # Torsion shear constant for tubular cross-section
        return self.Jxx / (1.28 * self.tf)

    @property
    def Syy(self):  # Bending modulus for tubular cross-section
        return self.BdgMyy

    @property
    def Szz(self):  # Bending modulus for tubular cross-section
        return self.BdgMzz

    @property
    def C(self):  # Torsion shear constant for tubular cross-section
        return self.TorsConst

    @property
    def Rgyr(self):  # Radius of Gyration for circular tube
        return np.sqrt(self.Jxx / self.Area)

    @property
    def CG(self):  # Radius of Gyration for circular tube
        return 0.5 * self.Hw + self.tf



class XlsxOperationException(Exception):
    """
    This exception is raised for errors that occur when processing .xlsx
    spreadsheet files.

    It has no custom implmentation. It is here to provide a class that
    can be specifically caught in an except statement.

    See also: https://docs.python.org/3/library/exceptions.html
    """

    pass



"""Provides the `Crane` class."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"

import simpy
from wisdem.orbit.core.defaults import process_times as pt
from wisdem.orbit.core.exceptions import ItemNotFound, CargoMassExceeded, DeckSpaceExceeded, InsufficientCable

# TODO: __str__ methods for Components


class Crane:
    """Base Crane Class"""

    def __init__(self, crane_specs):
        """
        Creates an instance of Crane.

        Parameters
        ----------
        crane_specs : dict
            Dictionary containing crane system specifications.
        """

        self.extract_crane_specs(crane_specs)

    def extract_crane_specs(self, crane_specs):
        """
        Extracts and defines crane specifications.

        Parameters
        ----------
        crane_specs : dict
            Dictionary of crane specifications.
        """

        # Operational Parameters
        self.max_lift = crane_specs.get("max_lift", None)
        self.max_hook_height = crane_specs.get("max_hook_height", None)
        self.max_windspeed = crane_specs.get("max_windspeed", 99)
        self._crane_rate = crane_specs.get("crane_rate", 100)

    def crane_rate(self, **kwargs):
        """Returns `self._crane_rate`."""

        return self._crane_rate

    @staticmethod
    def reequip(**kwargs):
        """
        Calculates time taken to change crane equipment.

        Parameters
        ----------
        crane_reequip_time : int | float
            Time required to change crane equipment (h).

        Returns
        -------
        reequip_time : float
            Time required to change crane equipment (h).
        """

        _key = "crane_reequip_time"
        duration = kwargs.get(_key, pt[_key])

        return duration


class DynamicPositioning:
    """Base Dynamic Positioning Class"""

    def __init__(self, dp_specs):
        """
        Creates an instance of DynamicPositioning.

        Parameters
        ----------
        dp_specs : dict
            Dictionary containing dynamic positioning specs.
        """

        self.extract_dp_specs(dp_specs)

    def extract_dp_specs(self, dp_specs):
        """
        Extracts and defines jacking system specifications.

        Parameters
        ----------
        jacksys_specs : dict
            Dictionary containing jacking system specifications.
        """

        self.dp_class = dp_specs.get("class", 1)


class JackingSys:
    """Base Jacking System Class"""

    def __init__(self, jacksys_specs):
        """
        Creates an instance of JackingSys.

        Parameters
        ----------
        jacksys_specs : dict
            Dictionary containing jacking system specifications.
        """

        self.extract_jacksys_specs(jacksys_specs)

    def extract_jacksys_specs(self, jacksys_specs):
        """
        Extracts and defines jacking system specifications.

        Parameters
        ----------
        jacksys_specs : dict
            Dictionary containing jacking system specifications.
        """

        # Physical Dimensions
        self.leg_length = jacksys_specs.get("leg_length", None)
        self.air_gap = jacksys_specs.get("air_gap", None)
        self.leg_pen = jacksys_specs.get("leg_pen", None)

        # Operational Parameters
        self.max_depth = jacksys_specs.get("max_depth", None)
        self.max_extension = jacksys_specs.get("max_extension", None)
        self.speed_below_depth = jacksys_specs.get("speed_below_depth", None)
        self.speed_above_depth = jacksys_specs.get("speed_above_depth", None)

    def jacking_time(self, extension, depth):
        """
        Calculates jacking time for a given depth.

        Parameters
        ----------
        extension : int | float
            Height to jack-up to or jack-down from (m).
        depth : int | float
            Depth at jack-up location (m).

        Returns
        -------
        extension_time : float
            Time required to jack-up to given extension (h).
        """

        if extension > self.max_extension:
            raise Exception("{} extension is greater than {} maximum" "".format(extension, self.max_extension))

        elif depth > self.max_depth:
            raise Exception("{} is beyond the operating depth {}" "".format(depth, self.max_depth))

        elif depth > extension:
            raise Exception("Extension must be greater than depth")

        else:
            return (depth / self.speed_below_depth + (extension - depth) / self.speed_above_depth) / 60


class VesselStorage(simpy.FilterStore):
    """Vessel Storage Class"""

    required_keys = ["type", "mass", "deck_space"]

    def __init__(self, env, max_cargo, max_deck_space, max_deck_load, **kwargs):
        """
        Creates an instance of VesselStorage.

        Parameters
        ----------
        env : simpy.Environment
            SimPy environment that simulation runs on.
        max_cargo : int | float
            Maximum mass the storage system can carry (t).
        max_deck_space : int | float
            Maximum deck space the storage system can use (m2).
        max_deck_load : int | float
            Maximum deck load that the storage system can apply (t/m2).
        """

        capacity = kwargs.get("capacity", float("inf"))
        super().__init__(env, capacity)

        self.max_cargo_mass = max_cargo
        self.max_deck_space = max_deck_space
        self.max_deck_load = max_deck_load

    @property
    def current_cargo_mass(self):
        """Returns current cargo mass in tons."""

        return sum([item.mass for item in self.items])

    @property
    def current_deck_space(self):
        """Returns current deck space used in m2."""

        return sum([item.deck_space for item in self.items])

    def put_item(self, item):
        """
        Checks VesselStorage specific constraints and triggers self.put()
        if successful.

        Items put into the instance should be a dictionary with the following
        attributes:
        - name
        - mass (t)
        - deck_space (m2)

        Parameters
        ----------
        item : dict
            Dictionary of item properties.
        """

        # if any(x not in item.keys() for x in self.required_keys):
        #     raise ItemPropertyNotDefined(item, self.required_keys)

        if self.current_deck_space + item.deck_space > self.max_deck_space:
            raise DeckSpaceExceeded(self.max_deck_space, self.current_deck_space, item)

        if self.current_cargo_mass + item.mass > self.max_cargo_mass:
            raise CargoMassExceeded(self.max_cargo_mass, self.current_cargo_mass, item)

        self.put(item)

    def get_item(self, _type):
        """
        Checks `self.items` for an item satisfying `item.type = _type`. Returns
        item if found, otherwise returns an error.

        Parameters
        ----------
        _type : str
            Type of item to retrieve.
        """

        target = None
        for i in self.items:
            if i.type == _type:
                target = i
                break

        if not target:
            raise ItemNotFound(_type)

        else:
            res = self.get(lambda x: x == target)
            return res.value

    def any_remaining(self, _type):
        """
        Checks `self.items` for an item satisfying `item.type = _type`. Returns
        True/False depending on if an item is found. Used to trigger vessel
        release if empty without having to wait for next self.get_item()
        iteration.

        Parameters
        ----------
        _type : str
            Type of item to retrieve.

        Returns
        -------
        resp : bool
            Indicates if any items in self.items satisfy `_type`.
        """

        target = None
        for i in self.items:
            if i.type == _type:
                target = i
                break

        if target:
            return True

        else:
            return False


class ScourProtectionStorage(simpy.Container):
    """Scour Protection Storage Class"""

    def __init__(self, env, max_mass, **kwargs):
        """
        Creates an instance of VesselStorage.

        Parameters
        ----------
        env : simpy.Environment
            SimPy environment that simulation runs on.
        max_mass : int | float
            Maximum mass the storage system can carry (t).
        """

        self.max_mass = max_mass
        super().__init__(env, self.max_mass)

    @property
    def available_capacity(self):
        """Returns available cargo capacity."""

        return self.max_mass - self.level


class CableCarousel(simpy.Container):
    """Cable Storage Class"""

    def __init__(self, env, max_mass, **kwargs):
        """
        Creates an instance of CableCarousel.

        Parameters
        ----------
        env : simpy.Environment
            SimPy environment that simulation runs on.
        max_mass : int | float
            Maximum mass the storage system can carry (t).
        """

        self.cable = None
        self.max_mass = max_mass
        super().__init__(env)

    @property
    def available_mass(self):
        """Returns available cargo mass capacity."""

        return self.max_mass - self.current_mass

    @property
    def current_mass(self):
        """Returns current cargo mass"""

        try:
            mass = self.level * self.cable.linear_density
            return mass

        except AttributeError:
            return 0

    def available_length(self, cable):
        """Returns available length capacity based on input linear density."""

        return self.available_mass / cable.linear_density

    def reset(self):
        """Resets `self.cable` and empties `self.level`."""

        if self.level != 0.0:
            _ = self.get(self.level)

        self.cable = None

    def load_cable(self, cable, length=None):
        """
        Loads input `cable` type onto `self.level`. If `length` isn't passed,
        defaults to maximum amount of cable that can be loaded.

        Parameters
        ----------
        cable : Cable | SimpleCable
        length : int | float

        Raises
        ------
        ValueError
        """

        if self.cable and self.cable != cable:
            raise AttributeError("Carousel already has a cable type.")

        self.cable = cable
        if length is None:
            # Load maximum amount
            length = self.available_length(self.cable)
            self.put(length)

        else:
            # Load length of cable
            proposed = length * cable.linear_density
            if proposed > self.available_mass:
                raise ValueError(f"Length {length} of {cable} can't be loaded.")

            self.put(length)

    def get_cable(self, length):
        """
        Retrieves `length` of cable from `self.level`.

        Parameters
        ----------
        length : int | float
            Length of cable to retrieve.

        Raises
        ------
        InsufficientCable
        """

        if self.cable is None:
            raise AttributeError("Carousel doesn't have any cable.")

        if length > self.level:
            raise InsufficientCable(self.level, length)

        else:
            return self.get(length).amount



"""Provides the base `Cable`, `Plant`, and `CableSystem` classes."""

__author__ = ["Matt Shields", "Rob Hammond"]
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Rob Hammond"
__email__ = "robert.hammond@nrel.gov"


import math
from collections import Counter, OrderedDict

import numpy as np
from scipy.optimize import fsolve
from wisdem.orbit.core.library import extract_library_specs
from wisdem.orbit.phases.design import DesignPhase


class Cable:
    """
    Base cable class

    Attributes
    ----------
    conductor_size : float
        Cable cross section in :math:`mm^2`.
    current_capacity : float
        Cable current rating at 1m burial depth, :math:`A`.
    rated_voltage : float
        Cable rated line-to-line voltage, :math:`kV`.
    ac_resistance : float
        Cable resistance for AC current, (ohms/km).
    inductance : float
        Cable inductance, :math:`\\frac{MHz}{km}`.
    capacitance : float
        Cable capacitance, :math:`\\frac{nF}{km}`.
    linear_density : float
        Dry mass per kilometer, :math:`\\frac{tonnes}{km}`.
    cost_per_km : int
        Cable cost per kilometer, :math:`\\frac{USD}{km}`.
    char_impedance : float
        Characteristic impedance of equivalent cable circuit, (ohms).
    power_factor : float
        Power factor of AC current in cable, no units.
    cable_power : float
        Maximum 3-phase power dissipated in cable in :math:`MW`.
    line_frequency: int, default: 60
        Frequency of the AC current, :math:`Hz`.
    """

    required = (
        "conductor_size",
        "current_capacity",
        "rated_voltage",
        "ac_resistance",
        "inductance",
        "capacitance",
        "linear_density",
        "cost_per_km",
        "name",
    )

    def __init__(self, cable_specs, **kwargs):
        """
        Create an instance of Cable (either array or export)

        Parameters
        ---------
        cable_specs : dict
            Dictionary containing cable specifications.
        kwargs : dict
            Additional user inputs.
        """

        needs_value = []
        for r in self.required:
            x = cable_specs.get(r, None)
            if x is not None:
                setattr(self, r, x)
            else:
                needs_value.append(r)

        if needs_value:
            raise ValueError(f"{needs_value} must be defined in cable_specs")

        self.line_frequency = cable_specs.get("line_frequency", 60)

        # Calc additional cable specs
        self.calc_char_impedance(**kwargs)
        self.calc_power_factor()
        self.calc_cable_power()

    def calc_char_impedance(self):
        """
        Calculate characteristic impedance of cable.
        """

        conductance = 1 / self.ac_resistance

        num = complex(
            self.ac_resistance,
            2 * math.pi * self.line_frequency * self.inductance,
        )
        den = complex(conductance, 2 * math.pi * self.line_frequency * self.capacitance)
        self.char_impedance = np.sqrt(num / den)

    def calc_power_factor(self):
        """
        Calculate power factor.
        """

        phase_angle = math.atan(np.imag(self.char_impedance) / np.real(self.char_impedance))
        self.power_factor = math.cos(phase_angle)

    def calc_cable_power(self):
        """
        Calculate maximum power transfer through 3-phase cable in :math:`MW`.
        """

        self.cable_power = np.sqrt(3) * self.rated_voltage * self.current_capacity * self.power_factor / 1000


class Plant:
    """
    A "data class" to create the windfarm specifications for
    `ArraySystemDesign`.

    Attributes
    ----------
    layout : str
        The layout of the windfarm. Can only be "grid", "ring", or "custom".
        ..note:: custom is not implemented at this time.
    num_turbines : int
        Number of turbines contained in the windfarm.
    site_depth : float
        Average depth at the site in km.
    turbine_rating : float
        Capacity of an individual turbine in MW.
    row_distance : float
        Distance between any two strings in a grid layout in km. This is not
        used for ring layouts or custom layouts.
    turbine_distance : float
        Distance between any two turbines in a string in km. This is not used
        for custom layouts.
    substation_distance : float
        The shortest distance between the offshore substation and the first
        turbine of each string in km. In the ring layout this distance is
        uniform across all strings. In grid layout this represents the
        perpendicular distance to the first row of turbines. This is not used
        in custom layouts.
    """

    expected_config = {
        "plant": {
            "num_turbines": "int",
            "layout": "str",
            "turbine_distance": "km (optional)",
            "turbine_spacing": "rotor diameters",
            "row_spacing": "rotor diameters (optional)",
            "row_distance": "km (optional)",
            "substation_distance": "km",
        },
        "site": {"depth": "m"},
        "turbine": {"turbine_rating": "MW", "rotor_diameter": "m"},
    }

    def __init__(self, config):
        """
        Creates the object.

        Parameters
        ----------
        config : dict
            Dictionary of configuration settings.
        """

        self.num_turbines = config["plant"]["num_turbines"]
        self.site_depth = config["site"]["depth"] / 1000.0
        self.turbine_rating = config["turbine"]["turbine_rating"]

        self.layout = config["plant"]["layout"].lower()
        if self.layout not in ("custom", "grid", "ring"):
            raise ValueError("config: site: layout should be one of " "'custom', 'ring', or 'grid'.")
        if self.layout != "custom":
            self._initialize_distances(config)

    def _initialize_distances(self, config):
        """
        Initializes the distance attributes that are created for a grid and
        ring layout.

        Parameters
        ----------
        config : dict
            Configuration dictionary.
        """

        rotor_diameter = config["turbine"]["rotor_diameter"]

        self.turbine_distance = config["plant"].get("turbine_distance", None)
        if self.turbine_distance is None:
            self.turbine_distance = rotor_diameter * config["plant"]["turbine_spacing"] / 1000.0

        self.substation_distance = config["plant"].get("substation_distance", None)
        if self.substation_distance is None:
            self.substation_distance = self.turbine_distance

        self.row_distance = config["plant"].get("row_distance", None)
        if self.row_distance is None:
            if self.layout == "grid":
                self.row_distance = rotor_diameter * config["plant"]["row_spacing"] / 1000.0
            else:
                self.row_distance = self.turbine_distance


class CableSystem(DesignPhase):
    """
    Base cabling system class. This is the parent class to
    `ArraySystemDesign` and `ExportSystemDesign`.

    Attributes
    ----------
    cable_type : str
        An input of "array" or "export" to signify which cabling system is
        being designed.
    cables : dict
        Dictionary of cables being used with items as {"cable_name": `Cable`}.

    Raises
    ------
    NotImplementedError
        The property `detailed_output` is not yet defined.
    Exception
        `cables` must be created in order to get outputs.
    """

    cables = None

    def __init__(self, config, cable_type, **kwargs):
        """
        Initializes the configuration.

        Parameters
        ----------
        cable_type : str
            Type of cabling system. Only "array" or "export" should be passed.
        """

        config = self.initialize_library(config, **kwargs)
        self.config = self.validate_config(config)

        self.cable_type = cable_type.lower()
        if self.cable_type not in ("array", "export"):
            raise ValueError(
                f"{self.cable_type} is invalid input to `cable_type`. '",
                "Input must be `array` or `export`.",
            )

        self._design = config["_".join((cable_type, "system_design"))]

    def _initialize_cables(self):
        """
        Creates the base cable objects for each type of array cable being used.
        """

        if isinstance(self._design["cables"], str):
            self._design["cables"] = [self._design["cables"]]
        if isinstance(self._design["cables"], list):
            _cables = {}
            for name in self._design["cables"]:
                _cables[name] = extract_library_specs("cables", name)
            self._design["cables"] = _cables

        _cables = {}
        for name, cable in self._design["cables"].items():
            cable.setdefault("name", name)
            _cables[name] = cable
        self._design["cables"] = _cables

        # Cables are ordered from smallest to largest
        cables = OrderedDict(
            sorted(
                self._design["cables"].items(),
                key=lambda item: item[1]["current_capacity"],
            )
        )

        # Instantiate cables as Cable objects. Use "name" property from file
        # instead of filename.
        self.cables = OrderedDict()
        for specs in cables.values():
            name = specs["name"]
            self.cables[name] = Cable(specs)

    def _get_touchdown_distance(self):
        """
        Returns the cable touchdown distance measured from the centerpoint of
        the substructure.

        If depth <= 60, default is 0km (straight down assumed for fixed bottom).
        If depth > 60, default is 0.3 * depth.
        """

        _design = f"{self.cable_type}_system_design"
        depth = self.config["site"]["depth"]
        touchdown = self.config[_design].get("touchdown_distance", None)

        if touchdown is not None:
            self.touchdown = touchdown

        else:
            if depth <= 60:
                self.touchdown = 0

            else:
                self.touchdown = depth * 0.3

    @staticmethod
    def _catenary(a, *data):
        """Simple catenary equation."""

        d, h = data
        res = a * np.cosh(h / a) - (d + a)
        return res

    def _get_catenary_length(self, d, h):
        """
        Returns the catenary length of a cable that touches down at depth `d`
        and horizontal distance `h`.

        Returns
        -------
        float
            Catenary length.
        """

        a = fsolve(self._catenary, 8, (d, h))

        x = np.linspace(0, h)
        y = a * np.cosh(x / a) - a

        if not np.isclose(y[-1], d):
            print("Warning: Catenary calculation failed. Reverting to simple vertical profile.")
            return d

        return np.trapz(np.sqrt(1 + np.gradient(y, x) ** 2), x)

    @property
    def free_cable_length(self):
        """Returns the length of the vertical portion of a cable section in km."""

        depth = self.config["site"]["depth"]

        if not self.touchdown:
            return depth / 1000

        return self._get_catenary_length(depth, self.touchdown) / 1000

    @property
    def cable_lengths_by_type(self):
        """
        Creates dictionary of lists of cable sections for each type of cable

        Returns
        -------
        lengths : dict
            A dictionary of the section lengths required for each type of cable
            to fully connect the array cabling system.
            E.g.: {`Cable`.`name`: np.ndarray(float)}
        """

        lengths = {name: self.sections_cable_lengths[np.where(self.sections_cables == name)] for name in self.cables}
        return lengths

    @property
    def total_cable_length_by_type(self):
        """
        Calculates the total cable length for each type of cable.

        Returns
        -------
        total : dict
            A dictionary of the total cable length for each type of cable.
            E.g.: {`Cable.name`: list(section_lengths)}
        """

        total = {name: sections.sum() for name, sections in self.cable_lengths_by_type.items()}
        return total

    @property
    def cost_by_type(self):
        """
        Calculates the cost of each array cable type.

        Returns
        -------
        cost : dict
            A dictionary of the total cost of each type of array cable.
            E.g.: {`Cable.name`: cost}
        """

        cost = {
            name: length * self.cables[name].cost_per_km for name, length in self.total_cable_length_by_type.items()
        }
        return cost

    @property
    def total_cost(self):
        """
        Calculates the cost of the array cabling system.

        Returns
        -------
        float
            Total cost of the array cabling system.
        """

        return sum(self.cost_by_type.values())

    @property
    def detailed_output(self):
        """Returns detailed design outputs."""

        _output = {
            "length": self.total_cable_length_by_type,
            "cost": self.cost_by_type,
        }

        return _output

    @property
    def design_result(self):
        """
        A dictionary of cables types and number of different cable lengths and
        linear density.

        Returns
        -------
        output : dict
            Dictionary of the number of section lengths and the linear density
            of each cable type.
             - <`cable_type`>_system: dict
                - cables: dict
                    - `Cable.name`: dict
                        - sections: [
                            (length of unique section, number of sections)
                          ],
                        - linear_density: `Cable.linear_density`
        """

        if self.cables is None:
            raise Exception(f"Has {self.__class__.__name__} been ran?")

        system = "_".join((self.cable_type, "system"))
        output = {system: {"cables": {}, "system_cost": self.total_cost}}
        _temp = output[system]["cables"]

        for name, cable in self.cables.items():
            try:
                sections = self.cable_lengths_by_type_speed[name]
            except AttributeError:
                sections = self.cable_lengths_by_type[name]

            try:
                sections = [(data[0], count, *data[1:]) for data, count in Counter(sections).items()]
            except IndexError:
                sections = [(*data[:-1], data[-1]) for data in Counter(sections).items()]

            if sections:
                _temp[name] = {
                    "cable_sections": sections,
                    "linear_density": cable.linear_density,
                }

        return output



__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from copy import deepcopy

from wisdem.orbit.phases import BasePhase

# Tests for key checking.
expected_config = {
    "param1": "str",
    "param2": {"param3": "int | float", "param4": "str (optional)"},
    "param5 (variable)": "int",
    "param6": {"param7": "float (optional)"},
}


def test_good_config():

    config = deepcopy(expected_config)
    missing = BasePhase._check_keys(expected_config, config)
    assert len(missing) == 0


def test_missing_key():

    config = deepcopy(expected_config)
    _ = config.pop("param1")
    missing = BasePhase._check_keys(expected_config, config)
    assert len(missing) == 1


def test_optional():

    config = deepcopy(expected_config)
    _ = config["param2"].pop("param4")
    missing = BasePhase._check_keys(expected_config, config)
    assert len(missing) == 0


def test_variable_key():

    config = deepcopy(expected_config)
    _ = config.pop("param5 (variable)")

    missing = BasePhase._check_keys(expected_config, config)
    assert len(missing) == 0


def test_optional_dict():

    config = deepcopy(expected_config)
    _ = config.pop("param6")

    missing = BasePhase._check_keys(expected_config, config)
    assert len(missing) == 0



import numpy as np
import openmdao.api as om
import wisdem.towerse.tower_props as tp
import wisdem.towerse.tower_struct as ts
import wisdem.commonse.utilization_constraints as util_con
from wisdem.towerse import NPTS_SOIL, get_nfull
from wisdem.commonse.environment import TowerSoil
from wisdem.commonse.cross_sections import CylindricalShellProperties
from wisdem.commonse.wind_wave_drag import CylinderEnvironment


class TowerLeanSE(om.Group):
    """
    This is a geometry preprocessing group for the tower.

    This group contains components that calculate the geometric properties of
    the tower, such as mass and moments of inertia, as well as geometric
    constraints like diameter-to-thickness and taper ratio. No static or dynamic
    analysis of the tower occurs here.

    """

    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        mod_opt = self.options["modeling_options"]["WISDEM"]["TowerSE"]

        n_height_tow = mod_opt["n_height_tower"]
        n_layers_tow = mod_opt["n_layers_tower"]
        n_height_mon = mod_opt["n_height_monopile"]
        n_layers_mon = mod_opt["n_layers_monopile"]
        if "n_height" in mod_opt:
            n_height = mod_opt["n_height"]
        else:
            n_height = mod_opt["n_height"] = n_height_tow if n_height_mon == 0 else n_height_tow + n_height_mon - 1
        nFull = get_nfull(n_height, nref=mod_opt["n_refine"])

        self.set_input_defaults("gravity_foundation_mass", 0.0, units="kg")
        self.set_input_defaults("transition_piece_mass", 0.0, units="kg")
        self.set_input_defaults("tower_outer_diameter", np.ones(n_height), units="m")
        self.set_input_defaults("tower_wall_thickness", np.ones(n_height), units="m")
        self.set_input_defaults("outfitting_factor", np.zeros(n_height - 1))
        self.set_input_defaults("water_depth", 0.0, units="m")
        self.set_input_defaults("hub_height", 0.0, units="m")
        self.set_input_defaults("rho", np.zeros(n_height - 1), units="kg/m**3")
        self.set_input_defaults("unit_cost", np.zeros(n_height - 1), units="USD/kg")
        self.set_input_defaults("labor_cost_rate", 0.0, units="USD/min")
        self.set_input_defaults("painting_cost_rate", 0.0, units="USD/m**2")

        # Inputs here are the outputs from the Tower component in load_IEA_yaml
        # TODO: Use reference axis and curvature, s, instead of assuming everything is vertical on z
        self.add_subsystem(
            "yaml",
            tp.DiscretizationYAML(
                n_height_tower=n_height_tow,
                n_height_monopile=n_height_mon,
                n_layers_tower=n_layers_tow,
                n_layers_monopile=n_layers_mon,
                n_mat=self.options["modeling_options"]["materials"]["n_mat"],
            ),
            promotes=["*"],
        )

        # Promote all but foundation_height so that we can override
        self.add_subsystem(
            "geometry",
            tp.CylinderDiscretization(nPoints=n_height),
            promotes=[
                "z_param",
                "z_full",
                "d_full",
                "t_full",
                ("section_height", "tower_section_height"),
                ("diameter", "tower_outer_diameter"),
                ("wall_thickness", "tower_wall_thickness"),
            ],
        )

        self.add_subsystem(
            "props", CylindricalShellProperties(nFull=nFull), promotes=["Az", "Asx", "Asy", "Ixx", "Iyy", "Jz"]
        )
        self.add_subsystem("tgeometry", tp.TowerDiscretization(n_height=n_height, n_refine=mod_opt["n_refine"]), promotes=["*"])

        self.add_subsystem(
            "cm",
            tp.CylinderMass(nPoints=nFull),
            promotes=["z_full", "d_full", "t_full", "labor_cost_rate", "painting_cost_rate"],
        )
        self.add_subsystem(
            "tm",
            tp.TowerMass(n_height=n_height,
                         n_refine=mod_opt["n_refine"]),
            promotes=[
                "z_full",
                "d_full",
                "tower_mass",
                "tower_center_of_mass",
                "tower_section_center_of_mass",
                "tower_I_base",
                "tower_cost",
                "gravity_foundation_mass",
                "gravity_foundation_I",
                "transition_piece_mass",
                "transition_piece_cost",
                "transition_piece_height",
                "transition_piece_I",
                "monopile_mass",
                "monopile_cost",
                "structural_mass",
                "structural_cost",
            ],
        )
        self.add_subsystem(
            "gc",
            util_con.GeometricConstraints(nPoints=n_height),
            promotes=[
                "constr_taper",
                "constr_d_to_t",
                "slope",
                ("d", "tower_outer_diameter"),
                ("t", "tower_wall_thickness"),
            ],
        )

        self.add_subsystem(
            "turb",
            tp.TurbineMass(),
            promotes=[
                "turbine_mass",
                "monopile_mass",
                "tower_mass",
                "tower_center_of_mass",
                "tower_I_base",
                "rna_mass",
                "rna_cg",
                "rna_I",
                "hub_height",
            ],
        )

        # Connections for geometry and mass
        self.connect("z_start", "geometry.foundation_height")
        self.connect("d_full", "props.d")
        self.connect("t_full", "props.t")
        self.connect("rho_full", "cm.rho")
        self.connect("outfitting_full", "cm.outfitting_factor")
        self.connect("unit_cost_full", "cm.material_cost_rate")
        self.connect("cm.mass", "tm.cylinder_mass")
        self.connect("cm.cost", "tm.cylinder_cost")
        self.connect("cm.center_of_mass", "tm.cylinder_center_of_mass")
        self.connect("cm.section_center_of_mass", "tm.cylinder_section_center_of_mass")
        self.connect("cm.I_base", "tm.cylinder_I_base")


class TowerSE(om.Group):
    """
    This is the main TowerSE group that performs analysis of the tower.

    This group takes in geometric inputs from TowerLeanSE and environmental and
    loading conditions.
    """

    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        mod_opt = self.options["modeling_options"]["WISDEM"]["TowerSE"]
        monopile = self.options["modeling_options"]["flags"]["monopile"]
        nLC = mod_opt["nLC"]  # not yet supported
        wind = mod_opt["wind"]  # not yet supported
        frame3dd_opt = mod_opt["frame3dd"]
        if "n_height" in mod_opt:
            n_height = mod_opt["n_height"]
        else:
            n_height_tow = mod_opt["n_height_tower"]
            n_height_mon = mod_opt["n_height_monopile"]
            n_height = mod_opt["n_height"] = n_height_tow if n_height_mon == 0 else n_height_tow + n_height_mon - 1
        nFull = get_nfull(n_height, nref=mod_opt["n_refine"])
        self.set_input_defaults("E", np.zeros(n_height - 1), units="N/m**2")
        self.set_input_defaults("G", np.zeros(n_height - 1), units="N/m**2")
        if monopile and mod_opt["soil_springs"]:
            self.set_input_defaults("G_soil", 0.0, units="N/m**2")
            self.set_input_defaults("nu_soil", 0.0)
        self.set_input_defaults("sigma_y", np.zeros(n_height - 1), units="N/m**2")
        self.set_input_defaults("life", 0.0)

        # Load baseline discretization
        self.add_subsystem("geom", TowerLeanSE(modeling_options=self.options["modeling_options"]), promotes=["*"])

        if monopile and mod_opt["soil_springs"]:
            self.add_subsystem(
                "soil",
                TowerSoil(npts=NPTS_SOIL),
                promotes=[("G", "G_soil"), ("nu", "nu_soil"), ("depth", "suctionpile_depth")],
            )
            self.connect("d_full", "soil.d0", src_indices=[0])

        # Add in all Components that drive load cases
        # Note multiple load cases have to be handled by replicating components and not groups/assemblies.
        # Replicating Groups replicates the IndepVarComps which doesn't play nicely in OpenMDAO
        prom = [("zref", "wind_reference_height"), "shearExp", "z0", "cd_usr", "yaw", "beta_wind", "rho_air", "mu_air"]
        if monopile:
            prom += [
                "beta_wave",
                "rho_water",
                "mu_water",
                "cm",
                "Uc",
                "Hsig_wave",
                "Tsig_wave",
                "water_depth",
            ]

        for iLC in range(nLC):
            lc = "" if nLC == 1 else str(iLC + 1)

            self.add_subsystem(
                "wind" + lc, CylinderEnvironment(nPoints=nFull, water_flag=monopile, wind=wind), promotes=prom
            )

            self.add_subsystem(
                "pre" + lc,
                ts.TowerPreFrame(
                    n_height=n_height,
                    monopile=monopile,
                    soil_springs=mod_opt["soil_springs"],
                    gravity_foundation=mod_opt["gravity_foundation"],
                    n_refine=mod_opt["n_refine"]
                ),
                promotes=[
                    "transition_piece_mass",
                    "transition_piece_height",
                    "transition_piece_I",
                    "gravity_foundation_mass",
                    "gravity_foundation_I",
                    "z_full",
                    "suctionpile_depth",
                ],
            )
            self.add_subsystem(
                "tower" + lc,
                ts.CylinderFrame3DD(
                    nFull=nFull,
                    nK=4 if monopile and not mod_opt["gravity_foundation"] else 1,
                    nMass=2,
                    nPL=1,
                    frame3dd_opt=frame3dd_opt,
                ),
                promotes=["Az", "Asx", "Asy", "Ixx", "Iyy", "Jz"],
            )
            self.add_subsystem(
                "post" + lc,
                ts.TowerPostFrame(modeling_options=mod_opt),
                promotes=[
                    "life",
                    "z_full",
                    "d_full",
                    "t_full",
                    "rho_full",
                    "E_full",
                    "G_full",
                    "sigma_y_full",
                    "suctionpile_depth",
                    "Az",
                    "Asx",
                    "Asy",
                    "Ixx",
                    "Iyy",
                    "Jz",
                ],
            )

            self.connect("z_full", ["wind" + lc + ".z", "tower" + lc + ".z"])
            self.connect("d_full", ["wind" + lc + ".d", "tower" + lc + ".d"])
            self.connect("t_full", "tower" + lc + ".t")

            self.connect("rho_full", "tower" + lc + ".rho")
            self.connect("E_full", "tower" + lc + ".E")
            self.connect("G_full", "tower" + lc + ".G")

            self.connect("pre" + lc + ".kidx", "tower" + lc + ".kidx")
            self.connect("pre" + lc + ".kx", "tower" + lc + ".kx")
            self.connect("pre" + lc + ".ky", "tower" + lc + ".ky")
            self.connect("pre" + lc + ".kz", "tower" + lc + ".kz")
            self.connect("pre" + lc + ".ktx", "tower" + lc + ".ktx")
            self.connect("pre" + lc + ".kty", "tower" + lc + ".kty")
            self.connect("pre" + lc + ".ktz", "tower" + lc + ".ktz")
            self.connect("pre" + lc + ".midx", "tower" + lc + ".midx")
            self.connect("pre" + lc + ".m", "tower" + lc + ".m")
            self.connect("pre" + lc + ".mIxx", "tower" + lc + ".mIxx")
            self.connect("pre" + lc + ".mIyy", "tower" + lc + ".mIyy")
            self.connect("pre" + lc + ".mIzz", "tower" + lc + ".mIzz")
            self.connect("pre" + lc + ".mIxy", "tower" + lc + ".mIxy")
            self.connect("pre" + lc + ".mIxz", "tower" + lc + ".mIxz")
            self.connect("pre" + lc + ".mIyz", "tower" + lc + ".mIyz")
            self.connect("pre" + lc + ".mrhox", "tower" + lc + ".mrhox")
            self.connect("pre" + lc + ".mrhoy", "tower" + lc + ".mrhoy")
            self.connect("pre" + lc + ".mrhoz", "tower" + lc + ".mrhoz")

            self.connect("pre" + lc + ".plidx", "tower" + lc + ".plidx")
            self.connect("pre" + lc + ".Fx", "tower" + lc + ".Fx")
            self.connect("pre" + lc + ".Fy", "tower" + lc + ".Fy")
            self.connect("pre" + lc + ".Fz", "tower" + lc + ".Fz")
            self.connect("pre" + lc + ".Mxx", "tower" + lc + ".Mxx")
            self.connect("pre" + lc + ".Myy", "tower" + lc + ".Myy")
            self.connect("pre" + lc + ".Mzz", "tower" + lc + ".Mzz")
            if monopile and mod_opt["soil_springs"]:
                self.connect("soil.z_k", "pre" + lc + ".z_soil")
                self.connect("soil.k", "pre" + lc + ".k_soil")

            self.connect("wind" + lc + ".Px", "tower" + lc + ".Px")
            self.connect("wind" + lc + ".Py", "tower" + lc + ".Py")
            self.connect("wind" + lc + ".Pz", "tower" + lc + ".Pz")

            self.connect("wind" + lc + ".qdyn", "post" + lc + ".qdyn")

            self.connect("tower" + lc + ".tower_Fz", "post" + lc + ".tower_Fz")
            self.connect("tower" + lc + ".tower_Vx", "post" + lc + ".tower_Vx")
            self.connect("tower" + lc + ".tower_Vy", "post" + lc + ".tower_Vy")
            self.connect("tower" + lc + ".tower_Mxx", "post" + lc + ".tower_Mxx")
            self.connect("tower" + lc + ".tower_Myy", "post" + lc + ".tower_Myy")
            self.connect("tower" + lc + ".tower_Mzz", "post" + lc + ".tower_Mzz")



"""Provides the base `DesignPhase` class."""

__author__ = ["Jake Nunemaker", "Rob Hammond"]
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from abc import abstractmethod

from wisdem.orbit.phases import BasePhase


class DesignPhase(BasePhase):
    """BasePhase subclass for design modules."""

    expected_config = None
    output_config = None

    @property
    @abstractmethod
    def design_result(self):
        """
        Returns result of DesignPhase to be passed into config and consumed by
        InstallPhase.

        Returns
        -------
        dict
            Dictionary of design results.
        """

        return {}



import os

import numpy as np
import jsonschema as json

try:
    import ruamel_yaml as ry
except:
    try:
        import ruamel.yaml as ry
    except:
        raise ImportError("No module named ruamel.yaml or ruamel_yaml")


fschema_geom = os.path.join(os.path.dirname(os.path.realpath(__file__)), "geometry_schema.yaml")
fschema_model = os.path.join(os.path.dirname(os.path.realpath(__file__)), "modeling_schema.yaml")
fschema_opt = os.path.join(os.path.dirname(os.path.realpath(__file__)), "analysis_schema.yaml")


# ---------------------
def load_yaml(fname_input):
    reader = ry.YAML(typ="safe", pure=True)
    with open(fname_input, "r", encoding="utf-8") as f:
        input_yaml = reader.load(f)
    return input_yaml


# ---------------------
def write_yaml(instance, foutput):
    # Write yaml with updated values
    yaml = ry.YAML()
    yaml.default_flow_style = None
    yaml.width = float("inf")
    yaml.indent(mapping=4, sequence=6, offset=3)
    yaml.allow_unicode = False
    with open(foutput, "w", encoding="utf-8") as f:
        yaml.dump(instance, f)


# ---------------------
# This is for when the defaults are in another file
def nested_get(indict, keylist):
    rv = indict
    for k in keylist:
        rv = rv[k]
    return rv


def nested_set(indict, keylist, val):
    rv = indict
    for k in keylist:
        if k == keylist[-1]:
            rv[k] = val
        else:
            rv = rv[k]


def integrate_defaults(instance, defaults, yaml_schema):
    # Prep iterative validator
    # json.validate(self.wt_init, yaml_schema)
    validator = json.Draft7Validator(yaml_schema)
    errors = validator.iter_errors(instance)

    # Loop over errors
    for e in errors:
        if e.validator == "required":
            for k in e.validator_value:
                if not k in e.instance.keys():
                    mypath = e.absolute_path.copy()
                    mypath.append(k)
                    v = nested_get(defaults, mypath)
                    if isinstance(v, dict) or isinstance(v, list) or v in ["name", "material"]:
                        # Too complicated to just copy over default, so give it back to the user
                        raise (e)
                    else:
                        print("WARNING: Missing value,", list(mypath), ", so setting to:", v)
                        nested_set(instance, mypath, v)
        else:
            raise (e)
    return instance


def simple_types(indict):
    rv = indict
    for k in rv.keys():
        if type(rv[k]) == type(np.array([])):
            rv[k] = rv[k].tolist()
        elif isinstance(rv[k], (float, int, list, dict, bool, str)):
            pass
        else:
            rv[k] = ""

        try:
            simple_types(rv[k])
        except:
            continue
    return rv


# ---------------------
# See: https://python-jsonschema.readthedocs.io/en/stable/faq/#why-doesn-t-my-schema-s-default-property-set-the-default-on-my-instance
def extend_with_default(validator_class):
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(validator, properties, instance, schema):
        for property, subschema in properties.items():
            if "default" in subschema:
                instance.setdefault(property, subschema["default"])

        for error in validate_properties(validator, properties, instance, schema):
            yield error

    return json.validators.extend(validator_class, {"properties": set_defaults})


DefaultValidatingDraft7Validator = extend_with_default(json.Draft7Validator)


def validate_without_defaults(finput, fschema):
    yaml_schema = load_yaml(fschema) if type(fschema) == type("") else fschema
    myobj = load_yaml(finput) if type(finput) == type("") else finput
    json.Draft7Validator(yaml_schema).validate(myobj)
    return myobj


def validate_with_defaults(finput, fschema):
    yaml_schema = load_yaml(fschema) if type(fschema) == type("") else fschema
    myobj = load_yaml(finput) if type(finput) == type("") else finput
    DefaultValidatingDraft7Validator(yaml_schema).validate(myobj)
    return myobj


# ---------------------
def load_geometry_yaml(finput):
    return validate_with_defaults(finput, fschema_geom)


def load_modeling_yaml(finput):
    return validate_with_defaults(finput, fschema_model)


def load_analysis_yaml(finput):
    return validate_with_defaults(finput, fschema_opt)


def write_geometry_yaml(instance, foutput):
    validate_without_defaults(instance, fschema_geom)
    sfx_str = ".yaml"
    if foutput[-5:] == sfx_str:
        sfx_str = ""
    write_yaml(instance, foutput + sfx_str)


def write_modeling_yaml(instance, foutput):
    validate_without_defaults(instance, fschema_model)
    sfx_str = ".yaml"
    if foutput[-5:] == sfx_str:
        foutput = foutput[-5:]
    elif foutput[-4:] == ".yml":
        foutput = foutput[-4:]
    sfx_str = "-modeling.yaml"

    instance2 = simple_types(instance)
    write_yaml(instance2, foutput + sfx_str)


def write_analysis_yaml(instance, foutput):
    validate_without_defaults(instance, fschema_opt)
    sfx_str = ".yaml"
    if foutput[-5:] == sfx_str:
        foutput = foutput[-5:]
    elif foutput[-4:] == ".yml":
        foutput = foutput[-4:]
    sfx_str = "-analysis.yaml"
    write_yaml(instance, foutput + sfx_str)


if __name__ == "__main__":
    yaml_schema = load_yaml(fschema_opt)
    myobj = load_yaml("sample_analysis.yaml")
    DefaultValidatingDraft7Validator(yaml_schema).validate(myobj)
    # validator.validate( myobj )
    print([k for k in myobj.keys()])
    print(myobj["general"])

    obj = {}
    schema = {"properties": {"foo": {"default": "bar"}}}
    # Note jsonschem.validate(obj, schema, cls=DefaultValidatingDraft7Validator)
    # will not work because the metaschema contains `default` directives.
    DefaultValidatingDraft7Validator(schema).validate(obj)
    print(obj)



"""Provides the `Vessel` class."""

__author__ = ["Jake Nunemaker", "Rob Hammond"]
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"

from math import ceil
from collections import Counter, namedtuple

import numpy as np
from marmot import Agent, le, process
from marmot._exceptions import StateExhausted, WindowNotFound, AgentNotRegistered
from wisdem.orbit.core.components import (
    Crane,
    JackingSys,
    CableCarousel,
    VesselStorage,
    DynamicPositioning,
    ScourProtectionStorage,
)
from wisdem.orbit.core.exceptions import ItemNotFound, MissingComponent

Trip = namedtuple("Trip", "cargo_mass deck_space items")


class Vessel(Agent):
    """Base Vessel Class"""

    def __init__(self, name, config, avail=1):
        """
        Creates an instance of Vessel.

        Parameters
        ----------
        specs : dict
            Nested dictionary containing vessel specifications.
        """

        super().__init__(name)
        self.config = config
        self.extract_vessel_dayrate()
        self.avail = avail

    def submit_action_log(self, action, duration, **kwargs):
        """
        Submits a log representing a completed `action` performed over time
        `duration`.

        This method overwrites the default `submit_action_log` in
        `marmot.Agent`, adding operation cost to every submitted log within
        ORBIT.

        Parameters
        ----------
        action : str
            Performed action.
        duration : int | float
            Duration of action.

        Raises
        ------
        AgentNotRegistered
        """

        if self.env is None:
            raise AgentNotRegistered(self)

        else:
            payload = {
                **kwargs,
                "agent": str(self),
                "action": action,
                "duration": float(duration),
                "cost": self.operation_cost(duration, **kwargs),
            }

            self.env._submit_log(payload, level="ACTION")

    @process
    def task_wrapper(self, name, duration, constraints={}, suspendable=False, **kwargs):

        duration /= self.avail
        yield self.task(name, duration, constraints, suspendable, **kwargs)

    def extract_vessel_dayrate(self):
        """
        Extracts the day rate of the vessel. If it isn't found, resorts to
        default values.
        """

        try:
            self.day_rate = self.config["vessel_specs"]["day_rate"]

        except KeyError:
            self.day_rate = np.NaN

    def mobilize(self):
        """
        Submits an action log representing the cost to mobilize the vessel at
        the start of an installation based on the vessel day rate.
        """

        days = self._vessel_specs.get("mobilization_days", 3)
        mult = self._vessel_specs.get("mobilization_mult", 0.5)

        self.submit_action_log("Mobilize", days * 24, cost_multiplier=mult)

    def operation_cost(self, hours, **kwargs):
        """
        Returns cost of an operation of duration `hours`.

        Parameters
        ----------
        hours : int | float
            Duration of operation in hours.
        mult : int | float
            Multiplier to use for the operation cost.
            Default: 1.
        """

        mult = kwargs.get("cost_multiplier", 1.0)
        return (self.day_rate / 24) * hours * mult

    @property
    def crane(self):
        """Returns configured `Crane` or raises `MissingComponent`."""
        try:
            return self._crane

        except AttributeError:
            raise MissingComponent(self, "Crane")

    @property
    def jacksys(self):
        """Returns configured `JackSys` or raises `MissingComponent`."""
        try:
            return self._jacksys

        except AttributeError:
            raise MissingComponent(self, "Jacking System")

    @property
    def dynamic_positioning(self):
        """Returns configured `DynamicPositioning` or raises `MissingComponent`."""
        try:
            return self._dp_system

        except AttributeError:
            raise MissingComponent(self, "Dynamic Positioning")

    @property
    def storage(self):
        """Returns configured `VesselStorage` or raises `MissingComponent`."""
        try:
            return self._storage

        except AttributeError:
            return MissingComponent(self, "Vessel Storage")

    @property
    def rock_storage(self):
        """Returns configured `ScourProtectionStorage` or raises `MissingComponent`."""
        try:
            return self._rock_storage

        except AttributeError:
            raise MissingComponent(self, "Scour Protection Storage")

    @property
    def cable_storage(self):
        """Returns configured `CableCarousel` or raises `MissingComponent`."""
        try:
            return self._cable_storage

        except AttributeError:
            raise MissingComponent(self, "Cable Storage")

    def initialize(self, mobilize=True):
        """
        Initializes vessel by extracting vessel/component specifications and
        running `self.mobilize()`.
        """

        self._vessel_specs = self.config.get("vessel_specs", {})
        self.extract_transport_specs()
        self.extract_jacksys_specs()
        self.extract_dp_specs()
        self.extract_crane_specs()
        self.extract_storage_specs()
        self.extract_cable_storage_specs()
        self.extract_scour_protection_specs()

        if mobilize:
            self.mobilize()

    def extract_transport_specs(self):
        """Extracts and defines transport related specifications."""

        self._transport_specs = self.config.get("transport_specs", {})
        self.transit_speed = self._transport_specs.get("transit_speed", None)

    def extract_crane_specs(self):
        """Extracts crane specifications if found."""

        self._crane_specs = self.config.get("crane_specs", {})
        if self._crane_specs:
            self._crane = Crane(self._crane_specs)

    def extract_jacksys_specs(self):
        """Extracts jacking system specifications if found."""

        self._jacksys_specs = self.config.get("jacksys_specs", {})
        if self._jacksys_specs:
            self._jacksys = JackingSys(self._jacksys_specs)

    def extract_dp_specs(self):
        """Extracts dynamic positioning system specifications if found."""

        self._dp_specs = self.config.get("dynamic_positioning_specs", {})
        if self._dp_specs:
            self._dp_system = DynamicPositioning(self._dp_specs)

    def extract_storage_specs(self):
        """Extracts storage system specifications if found."""

        self._storage_specs = self.config.get("storage_specs", {})
        if self._storage_specs:
            self.trip_data = []
            self._storage = VesselStorage(self.env, **self._storage_specs)

    def extract_cable_storage_specs(self):
        """Extracts and defines cable storage system specifications if found."""

        self._cable_storage_specs = self.config.get("cable_storage", {})
        if self._cable_storage_specs:
            self.trip_data = []
            self._cable_storage = CableCarousel(self.env, **self._cable_storage_specs)

    def extract_scour_protection_specs(self):
        """
        Extracts and defines scour protection installation specifications.

        Parameters
        ----------
        scour_protection_specs : dict
            Dictionary containing scour protection installation specifications.
        """

        self._sp_specs = self.config.get("spi_specs", {})

        _capacity = self._sp_specs.get("max_cargo_mass", None)
        if _capacity is None:
            capacity = self._storage_specs.get("max_cargo", None)

        else:
            capacity = _capacity

        if capacity:
            self._rock_storage = ScourProtectionStorage(self.env, capacity)

        self.scour_protection_install_speed = self._sp_specs.get("scour_protection_install_speed", 10)

    @process
    def get_item_from_storage(self, _type, vessel=None, release=False, **kwargs):
        """
        Retrieves an item which matches `item.type = _type` from `self.storage`
        or `vessel.storage` if configured.

        Parameters
        ----------
        _type : str
            Type of item to retrieve.
        vessel : Vessel | None
            Optional configuration to retrieve item from different vessel.
        release : bool
            If True, releases the vessel if it is empty.
        """

        if vessel is None:
            vessel = self

        try:
            item = vessel.storage.get_item(_type)

        except ItemNotFound as e:
            self.submit_debug_log(message="Item not found.")
            raise e

        action, time = item.release(**kwargs)
        if time > 0:
            yield self.task_wrapper(
                action,
                time,
                constraints=self.transit_limits,
                cost=self.operation_cost(time),
            )

        if release and vessel.storage.any_remaining(_type) is False:
            vessel.release.succeed()

        return item

    @process
    def transit(self, distance, **kwargs):
        """
        Generic transit task.

        Parameters
        ----------
        distance : int | float
            Distance to transit for.
        """

        time = self.transit_time(distance)
        yield self.task_wrapper(
            "Transit",
            time,
            constraints=self.transit_limits,
            suspendable=True,
            **kwargs,
        )

    def transit_time(self, distance):
        """
        Calculates transit time for a given distance.

        Parameters
        ----------
        distance : int | float
            Distance to travel (km).

        Returns
        -------
        transit_time : float
            Time required to travel 'distance' (h).
        """

        transit_time = distance / self.transit_speed

        return transit_time

    @property
    def transit_limits(self):
        """
        Returns dictionary of `marmot.Constraints` for 'windspeed' and
        'waveheight', representing the transit limits of the vessel.
        """

        _dict = {
            "windspeed": le(self._transport_specs["max_windspeed"]),
            "waveheight": le(self._transport_specs["max_waveheight"]),
        }

        return _dict

    @property
    def operational_limits(self):
        """
        Returns dictionary of `marmot.Constraints` for 'windspeed' and
        'waveheight', representing the operational limits of the vessel.
        """

        try:
            _ = getattr(self, "crane")
            max_windspeed = self._crane_specs["max_windspeed"]

        except MissingComponent:
            max_windspeed = self._transport_specs["max_windspeed"]

        _dict = {
            "windspeed": le(max_windspeed),
            "waveheight": le(self._transport_specs["max_waveheight"]),
        }

        return _dict

    def update_trip_data(self, cargo=True, deck=True, items=True):
        """
        Appends the current cargo utilization to the `self._cargo_utlization`.
        Used to collect cargo utilization statistics throughout a simulation.

        Parameters
        ----------
        items : bool
            Toggles optional item list collection.
        """

        storage = getattr(self, "storage", None)
        if storage is None:
            raise Exception("Vessel does not have storage capacity.")

        _cargo = storage.current_cargo_mass if cargo else np.NaN
        _deck = storage.current_deck_space if deck else np.NaN
        _items = dict(Counter(i for i in storage.items)) if items else np.NaN

        trip = Trip(cargo_mass=_cargo, deck_space=_deck, items=_items)

        self.trip_data.append(trip)

    @property
    def cargo_mass_list(self):
        """Returns cargo masss trips in self.trip_data."""

        return [trip.cargo_mass for trip in self.trip_data]

    @property
    def cargo_mass_utilizations(self):
        """Returns cargo mass utilizations for list of trips."""

        try:
            max_cargo_mass = self.storage.max_cargo_mass
            return np.array(self.cargo_mass_list) / max_cargo_mass

        except MissingComponent:
            return np.array(np.NaN)

    @property
    def deck_space_list(self):
        """Returns deck space used for trips in self.trip_data."""

        return [trip.deck_space for trip in self.trip_data]

    @property
    def deck_space_utilizations(self):
        """Returns deck space utilizations for list of trips."""

        try:
            max_deck_space = self.storage.max_deck_space
            return np.array(self.deck_space_list) / max_deck_space

        except MissingComponent:
            return np.array(np.NaN)

    @property
    def max_cargo_mass_utilization(self):
        """Returns maximum cargo mass utilization."""

        if not self.trip_data:
            return np.NaN

        return np.max(self.cargo_mass_utilizations)

    @property
    def min_cargo_mass_utilization(self):
        """Returns minimum cargo mass utilization."""

        if not self.trip_data:
            return np.NaN

        return np.min(self.cargo_mass_utilizations)

    @property
    def mean_cargo_mass_utilization(self):
        """Returns mean cargo mass utilization."""

        if not self.trip_data:
            return np.NaN

        return np.mean(self.cargo_mass_utilizations)

    @property
    def median_cargo_mass_utilization(self):
        """Returns median cargo mass utilization."""

        if not self.trip_data:
            return np.NaN

        return np.median(self.cargo_mass_utilizations)

    @property
    def max_deck_space_utilization(self):
        """Returns maximum deck_space utilization."""

        if not self.trip_data:
            return np.NaN

        return np.max(self.deck_space_utilizations)

    @property
    def min_deck_space_utilization(self):
        """Returns minimum deck_space utilization."""

        if not self.trip_data:
            return np.NaN

        return np.min(self.deck_space_utilizations)

    @property
    def mean_deck_space_utilization(self):
        """Returns mean deck space utilization."""

        if not self.trip_data:
            return np.NaN

        return np.mean(self.deck_space_utilizations)

    @property
    def median_deck_space_utilization(self):
        """Returns median deck space utilization."""

        if not self.trip_data:
            return np.NaN

        return np.median(self.deck_space_utilizations)

    @property
    def max_items_by_mass(self):
        """Returns items corresponding to `self.max_cargo_mass`."""

        if not self.trip_data:
            return np.NaN

        i = np.argmax(self.cargo_mass_list)
        return self.trip_data[i].items

    @property
    def min_items_by_mass(self):
        """Returns items corresponding to `self.min_cargo_mass`."""

        if not self.trip_data:
            return np.NaN

        i = np.argmin(self.cargo_mass_list)
        return self.trip_data[i].items

    @property
    def max_items_by_space(self):
        """Returns items corresponding to `self.max_deck_space`."""

        if not self.trip_data:
            return np.NaN

        i = np.argmax(self.deck_space_list)
        return self.trip_data[i].items

    @property
    def min_items_by_space(self):
        """Returns items corresponding to `self.min_deck_space`."""

        if not self.trip_data:
            return np.NaN

        i = np.argmin(self.deck_space_list)
        return self.trip_data[i].items



import cProfile
import pstats

import floatingse.floating as ff
import floatingse.semiInstance as se

# cProfile.run('ff.semiExample()','profout')
cProfile.run("se.psqp_optimal()", "profout")
p = pstats.Stats("profout")
n = 40
# Clean up filenames for the report
p.strip_dirs()

p.sort_stats("cumulative").print_stats(n)
p.sort_stats("time").print_stats(n)



"""Provides the `ScourProtectionDesign` class"""

__author__ = ["Rob Hammond", "Jake Nunemaker"]
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Rob Hammond"
__email__ = "robert.hammond@nrel.gov"

from math import ceil

import numpy as np
from wisdem.orbit.phases.design import DesignPhase


class ScourProtectionDesign(DesignPhase):
    """
    Calculates the necessary scour protection material for a fixed substructure.

    Parameters
    ----------
    config : dict
        Configuration dictionary with scour protection design inputs. See
        `expected_config` for expected data.

    Attributes
    ----------
    phi : float, default: 33.5
        Soil friction angle. Default is for medium density sand.
    equilibrium : float, default: 1.3
        Scour depth equilibrium, (S/D).
    rock_density : float, default: 2600
        Density of rocks used for scour protection in kg/(m^3).
    scour_depth : float
        Depth of the scour pit.
    protection_depth : float, default: 1m
        Depth of the scour protection.

    Other Attributes
    ----------------
    _design : dict
        Dictionary of specific scouring protection design parameters.
    num_turbines : int
        Number of turbines that need scouring protection.

    References
    ----------
    .. [1] Det Norske Veritas AS. (2014, May). Design of Offshore Wind Turbine
       Structures. Retrieved from
       https://rules.dnvgl.com/docs/pdf/DNV/codes/docs/2014-05/Os-J101.pdf
    """

    expected_config = {
        "monopile": {"diameter": "m"},
        "plant": {"num_turbines": "int"},
        "scour_protection_design": {
            "cost_per_tonne": "USD/t",
            "rock_density": "kg/m3 (optional)",
            "soil_friction_angle": "float (optional)",
            "scour_depth_equilibrium": "float (optional)",
            "scour_protection_depth": "m (optional)",
        },
    }

    output_config = {
        "scour_protection": {
            "tonnes_per_substructure": "t",
            "cost_per_tonne": "USD/t",
        }
    }

    def __init__(self, config, **kwargs):
        """
        Initialization function.

        Parameters
        ----------
        config : dict
            Configuration dictionary for scour protection design phase.
        """

        config = self.initialize_library(config, **kwargs)
        self.config = self.validate_config(config)
        self._design = self.config["scour_protection_design"]
        self.diameter = self.config["monopile"]["diameter"]
        self.num_turbines = self.config["plant"]["num_turbines"]

        self.phi = self._design.get("soil_friction_angle", 33.5)
        self.equilibrium = self._design.get("scour_depth_equilibrium", 1.3)
        self.rock_density = self._design.get("rock_density", 2600)
        self.protection_depth = self._design.get("scour_protection_depth", 1)

    def compute_scour_protection_tonnes_to_install(self):
        """
        Computes the amount of scour protection material that needs to be
        installed around a fixed substructure.

        Terms:
         * :math:`S =` Scour depth
         * :math:`D =` Monopile diameter
         * :math:`r =` Radius of scour protection from the center of the monopile
         * :math:`\\phi =` Soil friction angle

        Assumptions:
         * :math:`r = \\frac{D}{2} + \\frac{S}{\\tan(\\phi)}`

        References
        ----------
        .. [1] Det Norske Veritas AS. (2014, May). Design of Offshore Wind Turbine
        Structures. Retrieved from
        https://rules.dnvgl.com/docs/pdf/DNV/codes/docs/2014-05/Os-J101.pdf
        """

        self.scour_depth = self.equilibrium * self.diameter

        r = self.diameter / 2 + self.scour_depth / np.tan(np.radians(self.phi))

        volume = np.pi * self.protection_depth * (r ** 2 - (self.diameter / 2) ** 2)

        self.scour_protection_tonnes = ceil(self.rock_density * volume / 1000.0)

    def run(self):
        """
        Runs the required methods to be able to produce a `design_result`.
        """

        self.compute_scour_protection_tonnes_to_install()

    @property
    def total_cost(self):
        """Returns the total cost of the phase in $USD"""

        cost = self._design["cost_per_tonne"] * self.scour_protection_tonnes * self.num_turbines
        return cost

    @property
    def detailed_output(self):
        """Returns detailed outputs of the design."""

        _out = {
            "tonnes_per_substructure": self.scour_protection_tonnes,
            "cost_per_tonne": self._design["cost_per_tonne"],
        }
        return _out

    @property
    def design_result(self):
        """
        A dictionary of design results to passed to the scour protection
        installation simulation.

        Returns
        -------
        output : dict
             - ``scour_protection`` :`dict`
                - ``tonnes_per_substructure`` : `int`
        """

        output = {"scour_protection": self.detailed_output}
        return output



import unittest
import numpy as np
import numpy.testing as npt
from wisdem.pyframe3dd import Frame, Options, NodeData, ElementData, ReactionData, StaticLoadCase

class TestBreakdown(unittest.TestCase):
    
    def testAll(self):
        # Nodal data
        ntop = 5
        nbase = 10
        nall = ntop + nbase - 1
        xyz_top = np.c_[np.linspace(0.0, -10, ntop), np.zeros(ntop), 100*np.ones(ntop)]
        xyz_base = np.c_[np.zeros((nbase,2)), np.linspace(0, 100, nbase)]
        xyz_all = np.vstack( (xyz_base, xyz_top[1:,:]) )
        
        zero_top = np.zeros(ntop)
        inode_top = np.arange(ntop, dtype=np.int_)+1
        node_top = NodeData(inode_top, xyz_top[:,0], xyz_top[:,1], xyz_top[:,2], zero_top)
        
        zero_base = np.zeros(nbase)
        inode_base = np.arange(nbase, dtype=np.int_)+1
        node_base = NodeData(inode_base, xyz_base[:,0], xyz_base[:,1], xyz_base[:,2], zero_base)
        
        zero_all = np.zeros(nall)
        inode_all = np.arange(nall, dtype=np.int_)+1
        node_all = NodeData(inode_all, xyz_all[:,0], xyz_all[:,1], xyz_all[:,2], zero_all)

        # Reactions
        rnode = np.array([1], dtype=np.int_)
        Kval = np.array([1], dtype=np.int_)
        rigid = 1
        reactions = ReactionData(rnode, Kval, Kval, Kval, Kval, Kval, Kval, rigid)

        # Element data
        top1 = np.ones(ntop-1)
        base1 = np.ones(nbase-1)
        all1 = np.ones(nall-1)
        ielem_top = np.arange(1, ntop)
        ielem_base = np.arange(1, nbase)
        ielem_all = np.arange(1, nall)
        N1_top = np.arange(ntop-1, dtype=np.int_)+1
        N1_base = np.arange(nbase-1, dtype=np.int_)+1
        N1_all = np.arange(nall-1, dtype=np.int_)+1
        N2_top = N1_top + 1
        N2_base = N1_base + 1
        N2_all = N1_all + 1
        Ax = 75.0
        Asx = Asy = 50.0
        J0 = 200.0
        Ix = Iy = 100.0
        E = 5e6
        G = 5e6
        density = 8e3
        elem_top = ElementData(ielem_top, N1_top, N2_top,
                               Ax*top1, Asx*top1, Asy*top1,
                               J0*top1, Ix*top1, Iy*top1,
                               E*top1, G*top1, 0.0*top1, density*top1)
        elem_base = ElementData(ielem_base, N1_base, N2_base,
                               Ax*base1, Asx*base1, Asy*base1,
                               J0*base1, Ix*base1, Iy*base1,
                               E*base1, G*base1, 0.0*base1, density*base1)
        elem_all = ElementData(ielem_all, N1_all, N2_all,
                               Ax*all1, Asx*all1, Asy*all1,
                               J0*all1, Ix*all1, Iy*all1,
                               E*all1, G*all1, 0.0*all1, density*all1)

        # parameters
        shear = False  # 1: include shear deformation
        geom = False  # 1: include geometric stiffness
        dx = -1.0  # x-axis increment for internal forces
        options = Options(shear, geom, dx)

        frame_top = Frame(node_top, reactions, elem_top, options)
        frame_base = Frame(node_base, reactions, elem_base, options)
        frame_all = Frame(node_all, reactions, elem_all, options)

        # load case
        gx = 0.0
        gy = 0.0
        gz = -9.81
        load_top = StaticLoadCase(gx, gy, gz)
        load_base = StaticLoadCase(gx, gy, gz)
        load_all = StaticLoadCase(gx, gy, gz)

        # pseudo-rotor loads
        nF_top = [inode_top[-1]]
        nF_base = [inode_base[-1]]
        nF_all = [inode_all[-1]]
        F = 1e5
        M = 1e6
        load_top.changePointLoads(nF_top, [F], [F], [F], [M], [M], [M])
        load_all.changePointLoads(nF_all, [F], [F], [F], [M], [M], [M])

        frame_top.addLoadCase(load_top)
        frame_all.addLoadCase(load_all)

        # Added mass
        AN_top = [inode_top[-2]]
        AN_all = [inode_all[-2]]
        EMs = np.array([1e6])
        EMxx = EMyy = EMzz = EMxy = EMxz = EMyz = np.array([0.0])
        rhox = rhoy = rhoz = np.array([0.0])
        addGravityLoad = True
        frame_top.changeExtraNodeMass(AN_top, EMs, EMxx, EMyy, EMzz, EMxy, EMxz, EMyz, rhox, rhoy, rhoz, addGravityLoad)
        frame_all.changeExtraNodeMass(AN_all, EMs, EMxx, EMyy, EMzz, EMxy, EMxz, EMyz, rhox, rhoy, rhoz, addGravityLoad)

        # Run first models
        disp_top, forces_top, rxns_top, _, _, modal_top = frame_top.run()
        disp_all, forces_all, rxns_all, _, _, modal_all = frame_all.run()

        # Transfer loads to base
        load_base.changePointLoads(nF_base,
                                   [-rxns_top.Fx],
                                   [-rxns_top.Fy],
                                   [-rxns_top.Fz],
                                   [-rxns_top.Mxx],
                                   [-rxns_top.Myy],
                                   [-rxns_top.Mzz],
                                   )
        frame_base.addLoadCase(load_base)
        disp_base, forces_base, rxns_base, _, _, modal_base = frame_base.run()

        npt.assert_almost_equal(rxns_all.Fx, rxns_base.Fx, decimal=3)
        npt.assert_almost_equal(rxns_all.Fy, rxns_base.Fy, decimal=3)
        npt.assert_almost_equal(rxns_all.Fz, rxns_base.Fz, decimal=3)
        npt.assert_almost_equal(rxns_all.Mxx, rxns_base.Mxx, decimal=2)
        npt.assert_almost_equal(rxns_all.Myy, rxns_base.Myy, decimal=2)
        npt.assert_almost_equal(rxns_all.Mzz, rxns_base.Mzz, decimal=2)
        
def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestBreakdown))
    return suite


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)
    




import unittest

import numpy as np
import numpy.testing as npt
from wisdem.commonse.cross_sections import Tube, CylindricalShellProperties

npts = 100


class TestTube(unittest.TestCase):
    def setUp(self):
        self.params = {}
        self.unknowns = {}
        self.resid = None

        self.params["d"] = 2 * 5.0 * np.ones(npts)
        self.params["t"] = 1.0 * np.ones(npts - 1)

        self.mytube = CylindricalShellProperties(nFull=npts)

    def testTubeProperties(self):
        T = Tube(2 * 5.0, 1.0)

        self.assertAlmostEqual(T.Area, np.pi * 9.0)
        self.assertAlmostEqual(T.Jxx, np.pi * 369.0 / 4.0)
        self.assertAlmostEqual(T.Jyy, np.pi * 369.0 / 4.0)
        self.assertAlmostEqual(T.J0, np.pi * 369.0 / 2.0)
        self.assertAlmostEqual(T.S, np.pi * 369.0 / 4.0 / 5.0)
        self.assertAlmostEqual(T.C, np.pi * 369.0 / 2.0 / 5.0)

    def testOutputsIncremental(self):
        self.mytube.compute(self.params, self.unknowns)

        npt.assert_almost_equal(self.unknowns["Az"], np.pi * 9.0)
        npt.assert_almost_equal(self.unknowns["Ixx"], np.pi * 369.0 / 4.0)
        npt.assert_almost_equal(self.unknowns["Iyy"], np.pi * 369.0 / 4.0)
        npt.assert_almost_equal(self.unknowns["Jz"], np.pi * 369.0 / 2.0)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestTube))
    return suite


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



# MoorProps
# start of a script to hold the functions that hold the mooring line and anchor property info
# just simple functions for now, can import functions to other scripts if necessary
# very much in progress
# MoorProps started 4-26-21, getLineProps first initialized back in Sept/Oct 2020


import numpy as np
import wisdem.moorpy as mp


def getLineProps(dmm, type="chain", stud="studless", source="Orcaflex-altered", name=""):
    """getLineProps version 3.2: Restructuring v3.1 to 'Orcaflex-original' and 'Orcaflex-altered'

    Motivation: The existing public, and NREL-internal references for mooring line component property
    data are either proprietary, or unreliable and incomplete. The best way to derive new equations
    as a function of diameter is to use data from mooring line manufacturer's catalogs. Once developed,
    these new equations will serve as an updated version to compare against existing expressions.

    The goal is to have NREL's own library of mooring line property equations, but more research is needed.
    The original Orcaflex equations are the best we have right now and have been altered to include
    a quadratic chain MBL equation rather than a cubic, to avoid negative MBLs.
    Also, different cost models are put in to the altered version to avoid the Equimar data. Many sources
    researched for cost data, coefficients used are close to the old NREL internal data, but still an approximation.

    For more info, see the Mooring Component Properties Word doc.

    - This function requires at least one input: the line diameter in millimeters.
    - The rest of the inputs are optional: describe the desired type of line (chain, polyester, wire, etc.),
    the type of chain (studless or studlink), the source of data (Orcaflex-original or altered), or a name identifier
    - The function will output a MoorPy linetype object

    """

    if source == "Orcaflex-original":
        d = dmm / 1000  # orcaflex uses meters https://www.orcina.com/webhelp/OrcaFlex/

        if type == "chain":
            c = 1.96e4  # grade 2=1.37e4; grade 3=1.96e4; ORQ=2.11e4; R4=2.74e4
            MBL = c * d ** 2 * (44 - 80 * d) * 1000  # [N]  The same for both studless and studlink
            if stud == "studless":
                massden = 19.9 * d ** 2 * 1000  # [kg/m]
                EA = 0.854e8 * d ** 2 * 1000  # [N]
                d_vol = 1.8 * d  # [m]
            elif stud == "studlink" or stud == "stud":
                massden = 21.9 * d ** 2 * 1000  # [kg/m]
                EA = 1.010e8 * d ** 2 * 1000  # [N]
                d_vol = 1.89 * d ** 2  # [m]
            else:
                raise ValueError("getLineProps error: Choose either studless or stud chain type ")

        elif type == "nylon":
            massden = 0.6476 * d ** 2 * 1000  # [kg/m]
            EA = 1.18e5 * d ** 2 * 1000  # [N]
            MBL = 139357 * d ** 2 * 1000  # [N] for wet nylon line, 163950d^2 for dry nylon line
            d_vol = 0.85 * d  # [m]
        elif type == "polyester":
            massden = 0.7978 * d ** 2 * 1000  # [kg/m]
            EA = 1.09e6 * d ** 2 * 1000  # [N]
            MBL = 170466 * d ** 2 * 1000  # [N]
            d_vol = 0.86 * d  # [m]
        elif type == "polypropylene":
            massden = 0.4526 * d ** 2 * 1000  # [kg/m]
            EA = 1.06e6 * d ** 2 * 1000  # [N]
            MBL = 105990 * d ** 2 * 1000  # [N]
            d_vol = 0.80 * d  # [m]
        elif type == "wire-fiber" or type == "fiber":
            massden = 3.6109 * d ** 2 * 1000  # [kg/m]
            EA = 3.67e7 * d ** 2 * 1000  # [N]
            MBL = 584175 * d ** 2 * 1000  # [N]
            d_vol = 0.82 * d  # [m]
        elif type == "wire-wire" or type == "wire" or type == "IWRC":
            massden = 3.9897 * d ** 2 * 1000  # [kg/m]
            EA = 4.04e7 * d ** 2 * 1000  # [N]
            MBL = 633358 * d ** 2 * 1000  # [N]
            d_vol = 0.80 * d  # [m]
        else:
            raise ValueError("getLineProps error: Linetype not valid. Choose from given rope types or chain ")

        # cost
        # Derived from Equimar graph: https://tethys.pnnl.gov/sites/default/files/publications/EquiMar_D7.3.2.pdf
        if type == "chain":
            cost = (0.21 * (MBL / 9.81 / 1000)) * 1.29  # [$/m]
        elif type == "nylon" or type == "polyester" or type == "polypropylene":
            cost = (0.235 * (MBL / 9.81 / 1000)) * 1.29  # [$/m]
        elif type == "wire" or type == "wire-wire" or type == "IWRC" or type == "fiber" or type == "wire-fiber":
            cost = (0.18 * (MBL / 9.81 / 1000) + 90) * 1.29  # [$/m]
        else:
            raise ValueError("getLineProps error: Linetype not valid. Choose from given rope types or chain ")

    elif source == "Orcaflex-altered":
        d = dmm / 1000  # orcaflex uses meters https://www.orcina.com/webhelp/OrcaFlex/

        if type == "chain":
            c = 2.74e4  # grade 2=1.37e4; grade 3=1.96e4; ORQ=2.11e4; R4=2.74e4
            MBL = (
                (371360 * d ** 2 + 51382.72 * d) * (c / 2.11e4) * 1000
            )  # this is a fit quadratic term to the cubic MBL equation. No negatives
            if stud == "studless":
                massden = 19.9 * d ** 2 * 1000  # [kg/m]
                EA = 0.854e8 * d ** 2 * 1000  # [N]
                d_vol = 1.8 * d  # [m]
            elif stud == "studlink" or stud == "stud":
                massden = 21.9 * d ** 2 * 1000  # [kg/m]
                EA = 1.010e8 * d ** 2 * 1000  # [N]
                d_vol = 1.89 * d ** 2  # [m]
            else:
                raise ValueError("getLineProps error: Choose either studless or stud chain type ")

            # cost = 2.5*massden   # a ballpark for R4 chain
            # cost = (0.58*MBL/1000/9.81) - 87.6          # [$/m] from old NREL-internal
            # cost = 3.0*massden     # rough value similar to old NREL-internal
            cost = 2.585 * massden  # [($/kg)*(kg/m)=($/m)]
            # cost = 0.0

        elif type == "nylon":
            massden = 0.6476 * d ** 2 * 1000  # [kg/m]
            EA = 1.18e5 * d ** 2 * 1000  # [N]
            MBL = 139357 * d ** 2 * 1000  # [N] for wet nylon line, 163950d^2 for dry nylon line
            d_vol = 0.85 * d  # [m]
            cost = (0.42059603 * MBL / 1000 / 9.81) + 109.5  # [$/m] from old NREL-internal
        elif type == "polyester":
            massden = 0.7978 * d ** 2 * 1000  # [kg/m]
            EA = 1.09e6 * d ** 2 * 1000  # [N]
            MBL = 170466 * d ** 2 * 1000  # [N]
            d_vol = 0.86 * d  # [m]

            # cost = (0.42059603*MBL/1000/9.81) + 109.5   # [$/m] from old NREL-internal
            # cost = 1.1e-4*MBL               # rough value similar to old NREL-internal
            cost = 0.162 * (MBL / 9.81 / 1000)  # [$/m]

        elif type == "polypropylene":
            massden = 0.4526 * d ** 2 * 1000  # [kg/m]
            EA = 1.06e6 * d ** 2 * 1000  # [N]
            MBL = 105990 * d ** 2 * 1000  # [N]
            d_vol = 0.80 * d  # [m]
            cost = (0.42059603 * MBL / 1000 / 9.81) + 109.5  # [$/m] from old NREL-internal
        elif type == "wire-fiber" or type == "fiber":
            massden = 3.6109 * d ** 2 * 1000  # [kg/m]
            EA = 3.67e7 * d ** 2 * 1000  # [N]
            MBL = 584175 * d ** 2 * 1000  # [N]
            d_vol = 0.82 * d  # [m]
            cost = 0.53676471 * MBL / 1000 / 9.81  # [$/m] from old NREL-internal
        elif type == "wire-wire" or type == "wire" or type == "IWRC":
            massden = 3.9897 * d ** 2 * 1000  # [kg/m]
            EA = 4.04e7 * d ** 2 * 1000  # [N]
            MBL = 633358 * d ** 2 * 1000  # [N]
            d_vol = 0.80 * d  # [m]
            # cost = MBL * 900./15.0e6
            # cost = (0.33*MBL/1000/9.81) + 139.5         # [$/m] from old NREL-internal
            cost = 5.6e-5 * MBL  # rough value similar to old NREL-internal
        else:
            raise ValueError("getLineProps error: Linetype not valid. Choose from given rope types or chain ")

    elif source == "NREL":
        """
        getLineProps v3.1 used to have old NREL-internal equations here as a placeholder, but they were not trustworthy.
         - The chain equations used data from Vicinay which matched OrcaFlex data. The wire rope equations matched OrcaFlex well,
           the synthetic rope equations did not

        The idea is to have NREL's own library of mooring line property equations, but more research needs to be done.
        The 'OrcaFlex-altered' source version is a start and can change name to 'NREL' in the future, but it is
        still 'OrcaFlex-altered' because most of the equations are from OrcaFlex, which is the best we have right now.

        Future equations need to be optimization proof = no negative numbers anywhere (need to write an interpolation function)
        Can add new line types as well, such as Aramid or HMPE
        """
        pass

    # Set up a main identifier for the linetype. Useful for things like "chain_bot" or "chain_top"
    if name == "":
        typestring = f"{type}{dmm:.0f}"
    else:
        typestring = name

    notes = f"made with getLineProps - source: {source}"

    return mp.LineType(typestring, d_vol, massden, EA, MBL=MBL, cost=cost, notes=notes, input_type=type, input_d=dmm)


def getAnchorProps(fx, fz, type="drag-embedment", display=0):
    """ Calculates anchor required capacity and cost based on specified loadings and anchor type"""

    # for now this is based on API RP-2SK guidance for static analysis of permanent mooring systems
    # fx and fz are horizontal and vertical load components assumed to come from a dynamic (or equivalent) analysis.

    # mooring line tenLimit specified in yaml and inversed for a SF in constraints
    # take the line forces on the anchor and give 20% consideration for dynamic loads on lines
    # coefficients in front of fx and fz in each anchorType are the SF for that anchor for quasi-static (pages 30-31 of RP-2SK)

    # scale QS loads by 20% to approximate dynamic loads
    fx = 1.2 * fx
    fz = 1.2 * fz

    # note: capacity is measured here in kg force

    euros2dollars = 1.18  # the number of dollars there currently are in a euro (3-31-21)

    if type == "drag-embedment":
        capacity_x = 1.5 * fx / 9.81

        fzCost = 0
        # if fz > 0:
        #    fzCost = 1e3*fz
        #    if display > 0:  print('WARNING: Nonzero vertical load specified for a drag embedment anchor.')

        anchorMatCost = 0.188 * capacity_x + fzCost  # material cost
        anchorInstCost = 163548 * euros2dollars  # installation cost
        anchorDecomCost = 228967 * euros2dollars  # decommissioning cost

    elif type == "suction":
        capacity_x = 1.6 * fx / 9.81
        capacity_z = 2.0 * fz / 9.81
        capacity = np.linalg.norm([capacity_x, capacity_z])  # overall capacity, assuming in any direction for now
        anchorMatCost = 1.08 * capacity  # material cost
        anchorInstCost = 179331 * euros2dollars  # installation cost
        anchorDecomCost = 125532 * euros2dollars  # decommissioning cost

    elif type == "plate":
        capacity_x = 2.0 * fx / 9.81
        capacity_z = 2.0 * fz / 9.81
        capacity = np.linalg.norm([capacity_x, capacity_z])  # overall capacity, assuming in any direction for now
        raise ValueError("plate anchors not yet supported")

    elif type == "micropile":  # note: no API guidance on this one
        capacity_x = 2.0 * fx / 9.81
        capacity_z = 2.0 * fz / 9.81
        capacity = np.linalg.norm([capacity_x, capacity_z])  # overall capacity, assuming in any direction for now
        anchorMatCost = (
            200000 * 1.2 / 500000
        ) * capacity  # [(Euros*($/Euros)/kg)*kg] linear interpolation of material cost
        anchorInstCost = 0  # installation cost
        anchorDecomCost = 0  # decommissioning cost

    elif type == "SEPLA":  # cross between suction and plate
        capacity_x = 2.0 * fx / 9.81
        capacity_z = 2.0 * fz / 9.81
        capacity = np.linalg.norm([capacity_x, capacity_z])  # overall capacity, assuming in any direction for now
        anchorMatCost = 0.45 * capacity  # material cost
        anchorInstCost = 0  # installation cost
        anchorDecomCost = 0  # decommissioning cost

    else:
        raise ValueError(f"getAnchorProps received an unsupported anchor type ({type})")

    # mooring line sizing:  Tension limit for QS: 50% MBS.  Or FOS = 2

    return anchorMatCost, anchorInstCost, anchorDecomCost  # [USD]



"""
csmPPI.py

Created by George Scott on 2012-08-01.
Modified by Katherine Dykes 2012.
Copyright (c) NREL. All rights reserved.
"""
import os
import re


class Escalator:
    """
    cost escalator class - computes weighted sum of different PPI_tables
    """

    def __init__(self, x):
        """ x is a list """
        self.name = x[0]  # name
        self.tbls = x[1]  # list of table names
        self.wts = x[2]  # list of weights in pct (0-100.0)

        sum = 0.0
        for i in range(len(self.wts)):
            self.wts[i] *= 0.01
            sum += self.wts[i]
        if abs(sum - 1.0) > 0.0001:
            print(("Weights add up to {0:.5f}".format(sum)))

    def compute(self, ppitbls, sy, sm, ey, em):
        """
        returns cost escalator between start_yr/start_mon and end_yr/end_mon
        ppitbls is a dictionary of PPITbl objects, indexed by NAICS code
        """
        sum = 0.0
        for i in range(len(self.tbls)):
            key = self.tbls[i].strip()
            """if (key == 'GDP'):
                print('Skipping GDP')
                continue"""
            if key not in ppitbls:
                print(("No PPI table %s".format(key)))
                continue
            if (key == "326150P") and (self.name == "Advanced Blade material costs       "):
                ce = ppitbls[key].getEsc(2002, sm, ey, em)
            else:
                ce = ppitbls[key].getEsc(sy, sm, ey, em)
            sum += ce * self.wts[i]
        return sum


# ------------------------------------------------


class PPITbl:
    """ a PPITbl object represents a cost table """

    def __init__(self, code="", name="PPITbl"):
        self.cost = []  # empty list
        self.years = []
        self.name = name
        self.code = code
        pass

    def add_row(self, year, cost_array):
        self.cost.append(cost_array)
        self.years.append(year)
        return 1

    def getEsc(self, start_yr, start_mon, end_yr, end_mon, printFlag=0):
        """ return cost escalator between two dates (mon==13 is annual value) """
        start_row = start_yr - self.years[0]
        end_row = end_yr - self.years[0]

        if start_row < 0:
            print(("\n*** Year start_yr ${0:.2f} before table start {1}\n".format(start_yr, self.years[0])))
            return None
        if end_row >= len(self.cost):
            print(("\n*** Year end_yr ${0:.2f} after table end {1}\n".format(end_yr, self.years[-1])))
            return None
        if len(self.cost[start_row]) < start_mon:
            raise IndexError("Start_mon out of range")
        if len(self.cost[int(end_row)]) < int(end_mon):
            print(("\n*** EM {0:2d} > LER {0:2d} in table {}".format(end_mon, len(self.cost[end_row]), self.code)))
            raise IndexError("End_mon out of range")

        try:
            """print('SR {0:2d} ER {0:2d}'.format(start_row,end_row))
            print('LSR {0:2d} LER {0:2d}'.format(len(self.cost[start_row]),len(self.cost[end_row])))"""
            cost_start = self.cost[start_row][start_mon - 1]
            cost_end = self.cost[int(end_row)][int(end_mon) - 1]
        except:
            print(("Index out of range for table {} {}".format(self.code, self.name)))
            return None
        esc = cost_end / cost_start
        return esc


# --------------------------------------------------------------------------------------


class PPI:
    def __init__(self, ref_yr, ref_mon, curr_yr, curr_mon, debug=0):
        """
        Initialize the PPI class for calculation of PPI indices given a referene year/month and current year/month.

        Parameters
        ----------
        ref_yr : int
          reference PPI year
        ref_mon : int
          reference PPI month
        curr_yr : int
          current PPI year
        curr_mon : int
          current PPI month
        """

        # self.escData = [None] * 37
        self.escData = {}  # try a dictionary
        self.tblfile = (
            "PPI_Tables.txt"  # TODO: temporary solution - should update so it can locate it from dictionary etc
        )
        self.ppitbls = {}  # dictionary of PPITbl objects
        self.yrs_gdp = []
        self.ppi_gdp = []

        self.ref_yr = ref_yr
        self.ref_mon = ref_mon
        self.curr_yr = curr_yr
        self.curr_mon = curr_mon
        self.debug = debug

        fullfile = os.path.join(os.path.dirname(os.path.abspath(__file__)), self.tblfile)
        infile = open(fullfile, "r")  # infile = open(self.tblfile)

        itable = -1
        found_tables = False
        found_GDP = False
        iCode = {}
        code = ""

        for line in infile:
            words = line.split("\t")
            if not words:
                continue

            if words[0].startswith("Gross Domestic Product"):
                found_GDP = True
                continue
            if found_GDP and words[0].startswith("Year"):
                y = []
                for i in range(1, len(words)):
                    if words[i].startswith("20"):
                        y.append(int(words[i]))
                self.yrs_gdp = y
                continue
            if found_GDP and words[0].startswith("Absolute Value"):
                g = []
                num_re = re.compile(r"[\d\.]+")
                for i in range(1, len(words)):
                    if num_re.search(words[i]):
                        g.append(float(words[i]))
                self.ppi_gdp = g

                code = "GDP"
                self.ppitbls[code] = PPITbl(
                    code=code, name="Gross Domestic Product"
                )  # add a new element to self.ppitbls
                for i in range(len(y)):
                    rvals = []
                    for mon in range(13):
                        rvals.append(g[i])  # fill all monthly values with annual value
                    self.ppitbls[code].add_row(y[i], rvals)
                ippi = len(self.ppitbls) - 1
                iCode[code] = ippi  # index of tables by code
                if self.debug > 0:
                    print(("Created {0:2d} {} {}".format(ippi, code, self.ppitbls[code].name)))
                continue

            if words[0].startswith("NAICS"):
                found_tables = True
                itable += 1
                iyr = 0

                words[2] = words[2].replace(r'"', "")  # strip quotes from name
                code = words[1]
                self.ppitbls[code] = PPITbl(code=code, name=words[2])  # add a new element to self.ppitbls
                ippi = len(self.ppitbls) - 1
                if self.debug > 0:
                    print(("Created {0:2d} {} {}".format(ippi, code, self.ppitbls[code].name)))
                iCode[code] = ippi  # index of tables by code

            if found_tables and words[0].startswith("20"):  # a year number
                rvals = []
                i = 1
                while re.match(r"\d+\.", words[i]):
                    rvals.append(float(words[i]))
                    i += 1
                self.ppitbls[code].add_row(int(words[0]), rvals)
                iyr += 1

        self.escData["IPPI_BLD"] = Escalator(
            [
                "Baseline Blade material costs       ",
                ["3272123", "3255204", "332722489", "326150P"],
                [60.00, 23.00, 8.00, 9.00],
            ]
        )
        self.escData["IPPI_BLA"] = Escalator(
            [
                "Advanced Blade material costs       ",
                ["3272123", "3255204", "332722489", "326150P"],
                [61.00, 27.00, 3.00, 9.00],
            ]
        )
        self.escData["IPPI_BLL"] = Escalator(["Blade Labor costs                   ", ["GDP    "], [100.00]])
        self.escData["IPPI_HUB"] = Escalator(
            [
                "Hub                                 ",
                [
                    "3315113",
                ],
                [100.00],
            ]
        )
        self.escData["IPPI_PMB"] = Escalator(
            [
                "Pitch Mechanisms/Bearings           ",
                ["332991P", "3353123", "333612P  ", "334513"],
                [50.00, 20.00, 20.00, 10.00],
            ]
        )
        self.escData["IPPI_LSS"] = Escalator(["Low speed shaft                     ", ["3315131"], [100.00]])
        self.escData["IPPI_BRN"] = Escalator(["Bearings                            ", ["332991P"], [100.00]])
        self.escData["IPPI_GRB"] = Escalator(["Gearbox                             ", ["333612P"], [100.00]])
        self.escData["IPPI_BRK"] = Escalator(["Mech brake, HS cpling etc           ", ["3363401"], [100.00]])
        self.escData["IPPI_GEN"] = Escalator(["Generator                           ", ["335312P"], [100.00]])
        self.escData["IPPI_VSE"] = Escalator(["Variable spd electronics            ", ["335314P"], [100.00]])
        self.escData["IPPI_YAW"] = Escalator(
            [
                "Yaw drive & bearing                 ",
                [
                    "3353123",
                    "332991P",
                ],
                [50.00, 50.00],
            ]
        )
        self.escData["IPPI_MFM"] = Escalator(["Main frame                          ", ["3315113"], [100.00]])
        self.escData["IPPI_ELC"] = Escalator(
            [
                "Electrical connections              ",
                [
                    "335313P",
                    "3359291",
                    "GDP      ",
                ],
                [25.00, 60.00, 15.00],
            ]
        )
        self.escData["IPPI_HYD"] = Escalator(["Hydraulic system                    ", ["3339954"], [100.00]])
        self.escData["IPPI_NAC"] = Escalator(
            [
                "Nacelle                             ",
                [
                    "3272123",
                    "3255204",
                    "GDP      ",
                ],
                [55.00, 30.00, 15.00],
            ]
        )
        self.escData["IPPI_CTL"] = Escalator(["Control, safety system              ", ["334513 "], [100.00]])
        self.escData["IPPI_MAR"] = Escalator(["Marinization                        ", ["GDP    "], [100.00]])
        self.escData["IPPI_WAR"] = Escalator(["Offshore Warranty Premium           ", ["GDP    "], [100.00]])
        self.escData["IPPI_TWR"] = Escalator(["Tower                               ", ["331221 "], [100.00]])
        self.escData["IPPI_MPF"] = Escalator(["Monopole Foundations                ", ["BHVY   "], [100.00]])
        self.escData["IPPI_TPT"] = Escalator(["Transportation On/Offshore          ", ["4841212"], [100.00]])
        self.escData["IPPI_STP"] = Escalator(["Off Shore Site Prep                 ", ["BHVY   "], [100.00]])
        self.escData["IPPI_LAI"] = Escalator(["Land Based Assembly & installation  ", ["BHVY   "], [100.00]])
        self.escData["IPPI_OAI"] = Escalator(["Offshore Assembly & installation    ", ["BHVY   "], [100.00]])
        self.escData["IPPI_LEL"] = Escalator(
            [
                "Land Based Elect                    ",
                ["3353119", "335313P", "3359291  ", "GDP"],
                [40.00, 15.00, 35.00, 10.00],
            ]
        )
        self.escData["IPPI_OEL"] = Escalator(
            [
                "Offshore Elect                      ",
                ["3353119", "335313P", "3359291  ", "GDP"],
                [5.00, 5.00, 70.00, 20.00],
            ]
        )
        self.escData["IPPI_LPM"] = Escalator(["Permits, engineering (Land Based)   ", ["GDP    "], [100.00]])
        self.escData["IPPI_OPM"] = Escalator(["Permits, engineering (Offshore)     ", ["GDP    "], [100.00]])
        self.escData["IPPI_LLR"] = Escalator(["Land Based Levelized Replacement    ", ["GDP    "], [100.00]])
        self.escData["IPPI_OLR"] = Escalator(["Offshore Levelized Replacement      ", ["GDP    "], [100.00]])
        self.escData["IPPI_LOM"] = Escalator(["O&M Land Based                      ", ["GDP    "], [100.00]])
        self.escData["IPPI_OOM"] = Escalator(["O&M Offshore                        ", ["GDP    "], [100.00]])
        self.escData["IPPI_LSE"] = Escalator(["Land Based & Offshore Lease Cost    ", ["GDP    "], [100.00]])
        self.escData["IPPI_FND"] = Escalator(["Foundations                         ", ["BHVY   "], [100.00]])
        self.escData["IPPI_RDC"] = Escalator(["Road & Civil Work                   ", ["BHWY   "], [100.00]])
        self.escData["IPPI_PAE"] = Escalator(["Personnel Access Equipment          ", ["GDP    "], [100.00]])

    def compute(self, escCode, debug=0):
        """
        Returns the cost escalator for escData object 'escCode', using reference and current yr/mon values.

        Parameters
        ----------
        escCode : object
          codes for different PPI data

        Returns
        -------
        esc : float
          cost escalator for a particular PPI from reference year/month to current year month
        """
        # returns the cost escalator for escData object 'escCode', using reference and current yr/mon values

        if escCode not in self.escData:
            print(("Warning - no such code in PPI {0:.2f}".format(escCode)))
            # return 0
        esc = self.escData[escCode].compute(self.ppitbls, self.ref_yr, self.ref_mon, self.curr_yr, self.curr_mon)
        if debug > 0:
            print(
                (
                    "Escalator {} from {}{:02} to {}{:02} = {:.4}".format(
                        escCode, self.ref_yr, self.ref_mon, self.curr_yr, self.curr_mon, esc
                    )
                )
            )
        return esc


# --------------------------------------------------------------------------------------


def example():

    # test cases

    sy = 2002
    sm = 9
    ey = 2010
    em = 13
    ppi = PPI(sy, sm, ey, em)

    for i in list(ppi.escData.keys()):
        print((ppi.escData[i].name))
        # Use ppi.compute(code) to compute ppi for default interval
        print(("{0:2d} {:.4} ".format(i, ppi.compute(i))))
        print(" " "")


if __name__ == "__main__":

    example()



"""Provides the `SemiSubmersibleDesign` class (from OffshoreBOS)."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from wisdem.orbit.phases.design import DesignPhase


class SemiSubmersibleDesign(DesignPhase):
    """Semi-Submersible Substructure Design"""

    expected_config = {
        "site": {"depth": "m"},
        "plant": {"num_turbines": "int"},
        "turbine": {"turbine_rating": "MW"},
        "semisubmersible_design": {
            "stiffened_column_CR": "$/t (optional, default: 3120)",
            "truss_CR": "$/t (optional, default: 6250)",
            "heave_plate_CR": "$/t (optional, default: 6250)",
            "secondary_steel_CR": "$/t (optional, default: 7250)",
            "towing_speed": "km/h (optional, default: 6)",
        },
    }

    output_config = {
        "substructure": {
            "mass": "t",
            "unit_cost": "USD",
            "towing_speed": "km/h",
        }
    }

    def __init__(self, config, **kwargs):
        """
        Creates an instance of `SemiSubmersibleDesign`.

        Parameters
        ----------
        config : dict
        """

        config = self.initialize_library(config, **kwargs)
        self.config = self.validate_config(config)
        self._design = self.config.get("semisubmersible_design", {})

        self._outputs = {}

    def run(self):
        """Main run function."""

        substructure = {
            "mass": self.substructure_mass,
            "unit_cost": self.substructure_cost,
            "towing_speed": self._design.get("towing_speed", 6),
        }

        self._outputs["substructure"] = substructure

    @property
    def stiffened_column_mass(self):
        """
        Calculates the mass of the stiffened column for a single
        semi-submersible in tonnes. From original OffshoreBOS model.
        """

        rating = self.config["turbine"]["turbine_rating"]
        mass = -0.9581 * rating ** 2 + 40.89 * rating + 802.09

        return mass

    @property
    def stiffened_column_cost(self):
        """
        Calculates the cost of the stiffened column for a single
        semi-submersible. From original OffshoreBOS model.
        """

        cr = self._design.get("stiffened_column_CR", 3120)
        return self.stiffened_column_mass * cr

    @property
    def truss_mass(self):
        """
        Calculates the truss mass for a single semi-submersible in tonnes. From
        original OffshoreBOS model.
        """

        rating = self.config["turbine"]["turbine_rating"]
        mass = 2.7894 * rating ** 2 + 15.591 * rating + 266.03

        return mass

    @property
    def truss_cost(self):
        """
        Calculates the cost of the truss for a signle semi-submerisble. From
        original OffshoreBOS model.
        """

        cr = self._design.get("truss_CR", 6250)
        return self.truss_mass * cr

    @property
    def heave_plate_mass(self):
        """
        Calculates the heave plate mass for a single semi-submersible in tonnes.
        From original OffshoreBOS model.
        """

        rating = self.config["turbine"]["turbine_rating"]
        mass = -0.4397 * rating ** 2 + 21.545 * rating + 177.42

        return mass

    @property
    def heave_plate_cost(self):
        """
        Calculates the heave plate cost for a single semi-submersible. From
        original OffshoreBOS model.
        """

        cr = self._design.get("heave_plate_CR", 6250)
        return self.heave_plate_mass * cr

    @property
    def secondary_steel_mass(self):
        """
        Calculates the mass of the required secondary steel for a single
        semi-submersible. From original OffshoreBOS model.
        """

        rating = self.config["turbine"]["turbine_rating"]
        mass = -0.153 * rating ** 2 + 6.54 * rating + 128.34

        return mass

    @property
    def secondary_steel_cost(self):
        """
        Calculates the cost of the required secondary steel for a single
        semi-submersible. For original OffshoreBOS model.
        """

        cr = self._design.get("secondary_steel_CR", 7250)
        return self.secondary_steel_mass * cr

    @property
    def substructure_mass(self):
        """Returns single substructure mass."""

        return self.stiffened_column_mass + self.truss_mass + self.heave_plate_mass + self.secondary_steel_mass

    @property
    def substructure_cost(self):
        """Returns single substructure cost."""

        return self.stiffened_column_cost + self.truss_cost + self.heave_plate_cost + self.secondary_steel_cost

    @property
    def total_substructure_mass(self):
        """Returns mass of all substructures."""

        num = self.config["plant"]["num_turbines"]
        return num * self.substructure_mass

    @property
    def design_result(self):
        """Returns the result of `self.run()`"""

        if not self._outputs:
            raise Exception("Has `SemiSubmersibleDesign` been ran yet?")

        return self._outputs

    @property
    def total_cost(self):
        """Returns total phase cost in $USD."""

        num = self.config["plant"]["num_turbines"]
        return num * self.substructure_cost

    @property
    def detailed_output(self):
        """Returns detailed phase information."""

        _outputs = {
            "stiffened_column_mass": self.stiffened_column_mass,
            "stiffened_column_cost": self.stiffened_column_cost,
            "truss_mass": self.truss_mass,
            "truss_cost": self.truss_cost,
            "heave_plate_mass": self.heave_plate_mass,
            "heave_plate_cost": self.heave_plate_cost,
            "secondary_steel_mass": self.secondary_steel_mass,
            "secondary_steel_cost": self.secondary_steel_cost,
        }

        return _outputs



# make all classes and functions in MoorPy.py available under the main package
# from wisdem.moorpy.MoorPy import *

from wisdem.moorpy.body import Body

# new core MoorPy imports to eventually replace the above
from wisdem.moorpy.line import Line  # this will make mp.Line available from doing "import wisdem.moorpy as mp"
from wisdem.moorpy.point import Point
from wisdem.moorpy.system import System
from wisdem.moorpy.helpers import *
from wisdem.moorpy.Catenary import catenary
from wisdem.moorpy.lineType import LineType



import os
import unittest

import wisdem.glue_code.gc_LoadInputs as gcl

test_dir = (
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__)))))
    + os.sep
    + "examples"
    + os.sep
    + "02_reference_turbines"
    + os.sep
)


class TestLoadInputs(unittest.TestCase):
    def setUp(self):
        fname_wt_input = test_dir + "nrel5mw.yaml"
        fname_modeling_options = test_dir + "modeling_options.yaml"
        fname_analysis_options = test_dir + "analysis_options.yaml"

        self.myobj = gcl.WindTurbineOntologyPython(fname_wt_input, fname_modeling_options, fname_analysis_options)

    def testRunFlags(self):
        self.myobj.wt_init["airfoils"] = {}
        self.myobj.wt_init["components"]["blade"] = {}
        self.myobj.wt_init.pop("bos")
        self.myobj.wt_init["components"].pop("tower")
        self.myobj.set_run_flags()

        self.assertTrue(self.myobj.modeling_options["flags"]["airfoils"])
        self.assertTrue(self.myobj.modeling_options["flags"]["blade"])
        self.assertFalse(self.myobj.modeling_options["flags"]["bos"])
        self.assertFalse(self.myobj.modeling_options["flags"]["tower"])

    def testOptFlags(self):
        self.myobj.analysis_options["design_variables"]["test1"] = {}
        self.myobj.analysis_options["design_variables"]["test1"]["flag"] = False
        self.myobj.analysis_options["design_variables"]["test1"]["test2"] = {}
        self.myobj.analysis_options["design_variables"]["test1"]["test2"]["flag"] = False
        self.myobj.analysis_options["design_variables"]["test1"]["test2"]["test3"] = {}
        self.myobj.analysis_options["design_variables"]["test1"]["test2"]["test3"]["flag"] = False

        self.myobj.set_opt_flags()
        self.assertFalse(self.myobj.analysis_options["opt_flag"])

        self.myobj.analysis_options["design_variables"]["test1"]["flag"] = True
        self.myobj.set_opt_flags()
        self.assertTrue(self.myobj.analysis_options["opt_flag"])

        self.myobj.analysis_options["design_variables"]["test1"]["flag"] = False
        self.myobj.analysis_options["design_variables"]["test1"]["test2"]["test3"]["flag"] = True
        self.myobj.set_opt_flags()
        self.assertTrue(self.myobj.analysis_options["opt_flag"])

        self.myobj.analysis_options["design_variables"]["blade"]["aero_shape"]["twist"]["n_opt"] = 500
        self.myobj.analysis_options["design_variables"]["blade"]["aero_shape"]["chord"]["n_opt"] = 600
        self.myobj.analysis_options["design_variables"]["blade"]["structure"]["spar_cap_ss"]["n_opt"] = 700
        self.myobj.analysis_options["design_variables"]["blade"]["structure"]["spar_cap_ps"]["n_opt"] = 800
        self.myobj.set_opt_flags()
        self.assertEqual(
            self.myobj.analysis_options["design_variables"]["blade"]["aero_shape"]["twist"]["n_opt"],
            self.myobj.modeling_options["WISDEM"]["RotorSE"]["n_span"],
        )
        self.assertEqual(
            self.myobj.analysis_options["design_variables"]["blade"]["aero_shape"]["chord"]["n_opt"],
            self.myobj.modeling_options["WISDEM"]["RotorSE"]["n_span"],
        )
        self.assertEqual(
            self.myobj.analysis_options["design_variables"]["blade"]["structure"]["spar_cap_ss"]["n_opt"],
            self.myobj.modeling_options["WISDEM"]["RotorSE"]["n_span"],
        )
        self.assertEqual(
            self.myobj.analysis_options["design_variables"]["blade"]["structure"]["spar_cap_ps"]["n_opt"],
            self.myobj.modeling_options["WISDEM"]["RotorSE"]["n_span"],
        )


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestLoadInputs))
    return suite


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



import numpy as np


class Point:
    """A class for any object in the mooring system that can be described by three translational coorindates"""

    def __init__(self, mooringSys, num, type, r, m=0, v=0, fExt=np.zeros(3), DOFs=[0, 1, 2]):
        """Initialize Point attributes

        Parameters
        ----------
        mooringSys : system object
            The system object that contains the point object
        num : int
            indentifier number
        type : int
            the point type: 0 free to move, 1 fixed, -1 coupled externally
        r : array
            x,y,z coorindate position vector [m].
        m : float, optional
            mass [kg]. The default is 0.
        v : float, optional
            volume [m^3]. The default is 0.
        fExt : array, optional
            applied external force vector in global orientation (not including weight/buoyancy). The default is np.zeros(3).
        DOFs: list
            list of which coordinate directions are DOFs for this point (default 0,1,2=x,y,z). E.g. set [2] for vertical motion only.

        Returns
        -------
        None.

        """

        self.sys = mooringSys  # store a reference to the overall mooring system (instance of System class)

        self.number = num
        self.type = type  # 1: fixed/attached to something, 0 free to move, or -1 coupled externally
        self.r = np.array(r, dtype=np.float_)

        self.m = np.float_(m)
        self.v = np.float_(v)
        self.fExt = fExt  # external forces plus weight/buoyancy
        self.fBot = 10.0  # this is a seabed contact force that will be added if a point is specified below the seabed
        self.zSub = 0.0  # this is the depth that the point is positioned below the seabed (since r[2] will be capped at the depth)
        self.zTol = 2.0  # depth tolerance to be used when updating the point's position relative to the seabed

        self.DOFs = DOFs
        self.nDOF = len(DOFs)

        self.attached = []  # ID numbers of any Lines attached to the Point
        self.attachedEndB = []  # specifies which end of the line is attached (1: end B, 0: end A)

        # print("Created Point "+str(self.number))

    def attachLine(self, lineID, endB):
        """Adds a Line end to the Point

        Parameters
        ----------
        lineID : int
            The identifier ID number of a line
        endB : boolean
            Determines which end of the line is attached to the point

        Returns
        -------
        None.

        """

        self.attached.append(lineID)
        self.attachedEndB.append(endB)
        # print("attached Line "+str(lineID)+" to Point "+str(self.number))

    def detachLine(self, lineID, endB):
        """Detaches a Line end from the Point

        Parameters
        ----------
        lineID : int
            The identifier ID number of a line
        endB : boolean
            Determines which end of the line is to be detached from the point

        Returns
        -------
        None.

        """

        self.attached.pop(self.attached.index(lineID))
        self.attachedEndB.pop(self.attachedEndB.index(endB))
        print("detached Line " + str(lineID) + " from Point " + str(self.number))

    def setPosition(self, r):
        """Sets the position of the Point, along with that of any dependent objects.

        Parameters
        ----------
        r : array
            x,y,z coordinate position vector of the point [m]

        Raises
        ------
        ValueError
            If the length of the input r array is not of length 3

        Returns
        -------
        None.

        """

        # update the position of the Point itself
        if len(r) == 3:  # original case, setting all three coordinates as normal, asuming x,y,z
            self.r = np.array(r)
        elif len(r) == self.nDOF:
            self.r[
                self.DOFs
            ] = r  # this does a mapping based on self.DOFs, to support points with e.g. only a z DOF or only x and z DOFs
        else:
            raise ValueError(
                f"Point setPosition method requires an argument of size 3 or nDOF, but size {len(r):d} was provided"
            )

        # update the point's depth and position based on relation to seabed
        self.zSub = np.max([-self.zTol, -self.r[2] - self.sys.depth])  # depth of submergence in seabed if > -zTol
        self.r = np.array(
            [self.r[0], self.r[1], np.max([self.r[2], -self.sys.depth])]
        )  # don't let it sink below the seabed

        # update the position of any attached Line ends
        for LineID, endB in zip(self.attached, self.attachedEndB):
            self.sys.lineList[LineID - 1].setEndPosition(self.r, endB)

        if len(self.r) < 3:
            print("Double check how this point's position vector is calculated")
            breakpoint()

    def getForces(self, lines_only=False, seabed=True, xyz=False):
        """Sums the forces on the Point, including its own plus those of any attached Lines.

        Parameters
        ----------
        lines_only : boolean, optional
            An option for calculating forces from just the mooring lines or not. The default is False.
        seabed : bool, optional
            if False, will not include the effect of the seabed pushing the point up
        xyz : boolean, optional
            if False, returns only forces corresponding to enabled DOFs. If true, returns forces in x,y,z regardless of DOFs.

        Returns
        -------
        f : array
            The force vector applied to the point in its current position [N]

        """

        f = np.zeros(3)  # create empty force vector on the point

        if lines_only == False:

            f[2] += -self.m * self.sys.g  # add weight

            if self.r[2] < -1:  # add buoyancy if fully submerged
                f[2] += self.v * self.sys.rho * self.sys.g
            elif self.r[2] < 1:  # imagine a +/-1 m band at z=0 where buoyancy tapers to zero
                f[2] += self.v * self.sys.rho * self.sys.g * (0.5 - 0.5 * self.r[2])

            f += np.array(self.fExt)  # add external forces

            # handle case of Point resting on or below the seabed, to provide a restoring force
            # add smooth transition to fz=0 at seabed (starts at zTol above seabed)
            f[2] += max(self.m - self.v * self.sys.rho, 0) * self.sys.g * (self.zSub + self.zTol) / self.zTol

        # add forces from attached lines
        for LineID, endB in zip(self.attached, self.attachedEndB):
            f += self.sys.lineList[LineID - 1].getEndForce(endB)

        if xyz:
            return f
        else:
            return f[self.DOFs]  # return only the force(s) in the enable DOFs

    def getStiffness(self, X=[], tol=0.0001, dx=0.01):
        """Gets the stiffness matrix of the point due only to mooring lines with all other objects free to equilibrate.
        NOTE: This method currently isn't set up to worry about nDOF and DOFs settings of the Point. It only works for DOFs=[0,1,2].

        Parameters
        ----------
        X1 : array
            The position vector of the Point at which the stiffness matrix is to be calculated.
        dx : float, optional
            The change in displacement to be used for calculating the change in force. The default is 0.01.

        Returns
        -------
        K : matrix
            The stiffness matrix of the point at the given position X1.

        """

        # print("Getting Point "+str(self.number)+" stiffness matrix...")

        if len(X) == 3:
            X1 = np.array(X)
        elif len(X) == 0:
            X1 = self.r
        else:
            raise ValueError("Point.getStiffness expects the optional X parameter to be size 3")

        # set this Point's type to fixed so mooring system equilibrium response to its displacements can be found
        type0 = self.type  # store original type to restore later
        self.type = 1  # set type to 1 (not free) so that it won't be adjusted when finding equilibrium

        # if this Point is attached to a Body, set that Body's type to fixed so equilibrium can be found
        for body in self.sys.bodyList:  # search through all the bodies in the mooring system
            if self.number in body.attachedP:  # find the one that this Point is attached to (if at all)
                num = body.number  # store body number to index later
                Btype0 = body.type  # store original body type to restore later
                body.type = 1  # set body type to 1 (not free) so that it won't be adjusted when finding equilibrium

        # ensure this Point is positioned at the desired linearization point
        self.setPosition(X1)  # set position to linearization point
        self.sys.solveEquilibrium3(tol=tol)  # find equilibrium of mooring system given this Point in current position
        f = self.getForces(lines_only=True)  # get the net 6DOF forces/moments from any attached lines

        # Build a stiffness matrix by perturbing each DOF in turn
        K = np.zeros([3, 3])

        for i in range(len(K)):
            X2 = X1 + np.insert(
                np.zeros(2), i, dx
            )  # calculate perturbed Point position by adding dx to DOF in question
            self.setPosition(X2)  # perturb this Point's position
            self.sys.solveEquilibrium3(tol=tol)  # find equilibrium of mooring system given this Point's new position
            f_2 = self.getForces(lines_only=True)  # get the net 3DOF forces/moments from any attached lines

            K[i, :] = -(f_2 - f) / dx  # get stiffness in this DOF via finite difference and add to matrix column

        # ----------------- restore the system back to previous positions ------------------
        self.setPosition(X1)  # set position to linearization point
        self.sys.solveEquilibrium3(tol=tol)  # find equilibrium of mooring system given this Point in current position
        self.type = type0  # restore the Point's type to its original value
        for body in self.sys.bodyList:
            if self.number in body.attachedP:
                num = body.number
                self.sys.bodyList[
                    num - 1
                ].type = Btype0  # restore the type of the Body that the Point is attached to back to original value

        return K

    def getStiffnessA(self, lines_only=False, xyz=False):
        """Gets analytical stiffness matrix of Point due only to mooring lines with other objects fixed.

        Returns
        -------
        K : matrix
            3x3 analytic stiffness matrix.

        """

        # print("Getting Point "+str(self.number)+" analytic stiffness matrix...")

        K = np.zeros([3, 3])  # create an empty 3x3 stiffness matrix

        # append the stiffness matrix of each line attached to the point
        for lineID, endB in zip(self.attached, self.attachedEndB):
            line = self.sys.lineList[lineID - 1]
            K += line.getStiffnessMatrix()

        # NOTE: can rotate the line's stiffness matrix in either Line.getStiffnessMatrix() or here in Point.getStiffnessA()

        # add seabed or hydrostatic terms if needed
        if lines_only == False:

            # if within a +/-1 m band from z=0, apply a hydrostatic stiffness based on buoyancy
            if abs(self.r[2]) < 1:
                K[2, 2] += self.v * self.sys.rho * self.sys.g * 0.5

            # if less than zTol above the seabed (could even be below the seabed), apply a stiffness (should bring wet weight to zero at seabed)
            if self.r[2] < self.zTol - self.sys.depth:
                K[2, 2] += max(self.m - self.v * self.sys.rho, 0) * self.sys.g / self.zTol

        if xyz:  # if asked to output all DOFs, do it
            return K
        else:  # otherwise only return rows/columns of active DOFs
            return K[:, self.DOFs][self.DOFs, :]



import unittest

import openmdao.api as om
import wisdem.nrelcsm.nrel_csm_mass_2015 as nct2015


class TestAll(unittest.TestCase):
    def testMass(self):

        # simple test of module
        prob = om.Problem()
        prob.model = nct2015.nrel_csm_mass_2015()
        prob.setup()

        prob["rotor_diameter"] = 126.0
        prob["turbine_class"] = 1
        prob["blade_has_carbon"] = False
        prob["blade_number"] = 3
        prob["machine_rating"] = 5000.0
        prob["hub_height"] = 90.0
        prob["main_bearing_number"] = 2
        prob["crane"] = True
        prob["max_tip_speed"] = 80.0
        prob["max_efficiency"] = 0.90

        prob.run_model()

        self.assertAlmostEqual(float(prob["blade_mass"]), 18590.66820649, 2)
        self.assertAlmostEqual(float(prob["hub_mass"]), 44078.53687493, 2)
        self.assertAlmostEqual(float(prob["pitch_system_mass"]), 10798.90594644, 2)
        self.assertAlmostEqual(float(prob["spinner_mass"]), 973.0, 2)
        self.assertAlmostEqual(float(prob["lss_mass"]), 22820.27928238, 2)
        self.assertAlmostEqual(float(prob["main_bearing_mass"]), 2245.41649102, 2)
        self.assertAlmostEqual(float(prob["rated_rpm"]), 12.1260909, 2)
        self.assertAlmostEqual(float(prob["rotor_torque"]), 4375000.0, 2)
        self.assertAlmostEqual(float(prob["gearbox_mass"]), 43468.32086769, 2)
        self.assertAlmostEqual(float(prob["hss_mass"]), 994.7, 2)
        self.assertAlmostEqual(float(prob["generator_mass"]), 14900.0, 2)
        self.assertAlmostEqual(float(prob["bedplate_mass"]), 41765.26095285, 2)
        self.assertAlmostEqual(float(prob["yaw_mass"]), 12329.96247921, 2)
        self.assertAlmostEqual(float(prob["hvac_mass"]), 400.0, 2)
        self.assertAlmostEqual(float(prob["cover_mass"]), 6836.69, 2)
        self.assertAlmostEqual(float(prob["platforms_mass"]), 8220.65761911, 2)
        self.assertAlmostEqual(float(prob["transformer_mass"]), 11485.0, 2)
        self.assertAlmostEqual(float(prob["tower_mass"]), 182336.48057717, 2)
        self.assertAlmostEqual(float(prob["hub_system_mass"]), 55850.44282136, 2)
        self.assertAlmostEqual(float(prob["rotor_mass"]), 111622.44744083, 2)
        self.assertAlmostEqual(float(prob["nacelle_mass"]), 173049.20418327392, 2)
        self.assertAlmostEqual(float(prob["turbine_mass"]), 467008.1322012795, 2)

    def testMassAndCost(self):

        # simple test of module
        prob = om.Problem()
        prob.model = nct2015.nrel_csm_2015()
        prob.setup()

        prob["rotor_diameter"] = 126.0
        prob["turbine_class"] = 1
        prob["blade_has_carbon"] = False
        prob["blade_number"] = 3
        prob["machine_rating"] = 5000.0
        prob["hub_height"] = 90.0
        prob["main_bearing_number"] = 2
        prob["crane"] = True
        prob["max_tip_speed"] = 80.0
        prob["max_efficiency"] = 0.90

        prob.run_model()

        self.assertAlmostEqual(float(prob["blade_mass"]), 18590.66820649, 2)
        self.assertAlmostEqual(float(prob["hub_mass"]), 44078.53687493, 2)
        self.assertAlmostEqual(float(prob["pitch_system_mass"]), 10798.90594644, 2)
        self.assertAlmostEqual(float(prob["spinner_mass"]), 973.0, 2)
        self.assertAlmostEqual(float(prob["lss_mass"]), 22820.27928238, 2)
        self.assertAlmostEqual(float(prob["main_bearing_mass"]), 2245.41649102, 2)
        self.assertAlmostEqual(float(prob["rated_rpm"]), 12.1260909, 2)
        self.assertAlmostEqual(float(prob["rotor_torque"]), 4375000.0, 2)
        self.assertAlmostEqual(float(prob["gearbox_mass"]), 43468.32086769, 2)
        self.assertAlmostEqual(float(prob["hss_mass"]), 994.7, 2)
        self.assertAlmostEqual(float(prob["generator_mass"]), 14900.0, 2)
        self.assertAlmostEqual(float(prob["bedplate_mass"]), 41765.26095285, 2)
        self.assertAlmostEqual(float(prob["yaw_mass"]), 12329.96247921, 2)
        self.assertAlmostEqual(float(prob["hvac_mass"]), 400.0, 2)
        self.assertAlmostEqual(float(prob["cover_mass"]), 6836.69, 2)
        self.assertAlmostEqual(float(prob["platforms_mass"]), 8220.65761911, 2)
        self.assertAlmostEqual(float(prob["transformer_mass"]), 11485.0, 2)
        self.assertAlmostEqual(float(prob["tower_mass"]), 182336.48057717, 2)
        self.assertAlmostEqual(float(prob["hub_system_mass"]), 55850.44282136, 2)
        self.assertAlmostEqual(float(prob["rotor_mass"]), 111622.44744083, 2)
        self.assertAlmostEqual(float(prob["nacelle_mass"]), 173049.20418327392, 2)
        self.assertAlmostEqual(float(prob["turbine_mass"]), 467008.1322012795, 2)

        self.assertAlmostEqual(float(prob["blade_cost"]), 271423.75581475, 2)
        self.assertAlmostEqual(float(prob["hub_cost"]), 171906.29381221, 2)
        self.assertAlmostEqual(float(prob["pitch_system_cost"]), 238655.82141628, 2)
        self.assertAlmostEqual(float(prob["spinner_cost"]), 10800.3, 2)
        self.assertAlmostEqual(float(prob["hub_system_mass_tcc"]), 55850.44282136, 2)
        self.assertAlmostEqual(float(prob["hub_system_cost"]), 421362.41522849, 2)
        self.assertAlmostEqual(float(prob["rotor_cost"]), 1235633.68267274, 2)
        self.assertAlmostEqual(float(prob["rotor_mass_tcc"]), 111622.44744083, 2)
        self.assertAlmostEqual(float(prob["lss_cost"]), 271561.32346034, 2)
        self.assertAlmostEqual(float(prob["main_bearing_cost"]), 10104.37420958, 2)
        self.assertAlmostEqual(float(prob["gearbox_cost"]), 560741.33919321, 2)
        self.assertAlmostEqual(float(prob["hss_cost"]), 6763.96, 2)
        self.assertAlmostEqual(float(prob["generator_cost"]), 184760.0, 2)
        self.assertAlmostEqual(float(prob["bedplate_cost"]), 121119.25676326, 2)
        self.assertAlmostEqual(float(prob["yaw_system_cost"]), 102338.68857747, 2)
        self.assertAlmostEqual(float(prob["hvac_cost"]), 49600.0, 2)
        self.assertAlmostEqual(float(prob["controls_cost"]), 105750.0, 2)
        self.assertAlmostEqual(float(prob["converter_cost"]), 0.0, 2)
        self.assertAlmostEqual(float(prob["elec_cost"]), 209250.0, 2)
        self.assertAlmostEqual(float(prob["cover_cost"]), 38969.133, 2)
        self.assertAlmostEqual(float(prob["platforms_cost"]), 101273.24528671, 2)
        self.assertAlmostEqual(float(prob["transformer_cost"]), 215918.0, 2)
        self.assertAlmostEqual(float(prob["nacelle_cost"]), 2007604.267200145, 2)
        self.assertAlmostEqual(float(prob["nacelle_mass_tcc"]), 173049.20418327392, 2)
        self.assertAlmostEqual(float(prob["tower_parts_cost"]), 528775.7936738, 2)
        self.assertAlmostEqual(float(prob["tower_cost"]), 528775.7936738, 2)
        self.assertAlmostEqual(float(prob["turbine_mass_tcc"]), 467008.1322012795, 2)
        self.assertAlmostEqual(float(prob["turbine_cost"]), 3772013.7435466847, 2)
        self.assertAlmostEqual(float(prob["turbine_cost_kW"]), 754.402748709337, 2)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestAll))
    return suite


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



import unittest

import numpy as np
import numpy.testing as npt
import wisdem.commonse.frustum as f
from wisdem.commonse import eps

myones = np.ones((100,))
rb = 4.0
rt = 2.0
t = 0.1
h = 3.0


class TestFrustum(unittest.TestCase):
    def testFrustumVol(self):

        V = np.pi / 3 * h * (rb ** 2 + rt ** 2 + rb * rt)

        # Test volume- scalar and vector inputs
        self.assertEqual(f.frustumVol(rb, rt, h, False), V)
        self.assertEqual(f.frustumVol(2 * rb, 2 * rt, h, True), V)
        npt.assert_equal(f.frustumVol(rb * myones, rt * myones, h * myones, False), V * myones)
        npt.assert_equal(f.frustumVol(2 * rb * myones, 2 * rt * myones, h * myones, True), V * myones)

    def testFrustumCG_solid(self):

        cg_solid = h / 4 * (rb ** 2 + 3 * rt ** 2 + 2 * rb * rt) / (rb ** 2 + rt ** 2 + rb * rt)

        # Test cg of solid- scalar and vector inputs
        self.assertEqual(f.frustumCG(rb, rt, h, False), cg_solid)
        self.assertEqual(f.frustumCG(2 * rb, 2 * rt, h, True), cg_solid)
        npt.assert_equal(f.frustumCG(rb * myones, rt * myones, h * myones, False), cg_solid * myones)
        npt.assert_equal(f.frustumCG(2 * rb * myones, 2 * rt * myones, h * myones, True), cg_solid * myones)

    def testFrustum_shell(self):

        # In limit of thickness approaching radius, should recover regular formulas
        self.assertEqual(f.frustumShellVol(rb, rb, rb, h, False), f.frustumVol(rb, rb, h, False))
        self.assertEqual(f.frustumShellVol(2 * rt, 2 * rt, rt, h, True), f.frustumVol(rt, rt, h, False))

        self.assertEqual(f.frustumShellCG(rb, rb, rb, h, False), f.frustumCG(rb, rb, h, False))
        self.assertEqual(f.frustumShellCG(2 * rt, 2 * rt, rt, h, True), f.frustumCG(rt, rt, h, False))

        self.assertEqual(f.frustumShellIzz(rb, rb, rb, h, False), f.frustumIzz(rb, rb, h, False))
        self.assertEqual(f.frustumShellIzz(2 * rt, 2 * rt, rt, h, True), f.frustumIzz(rt, rt, h, False))

        self.assertAlmostEqual(f.frustumShellIxx(rb, rb, rb - eps, h, False), f.frustumIxx(rb, rb, h, False))
        self.assertAlmostEqual(f.frustumShellIxx(2 * rt, 2 * rt, rt - eps, h, True), f.frustumIxx(rt, rt, h, False))


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestFrustum))
    return suite


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)
